import numpy

def get_direction_increment(
        directions_radians: numpy.ndarray) -> numpy.ndarray:
    """
    calculate the stepsize used for midpoint integration. The directions
    represent the center of the interval - and we want to find the dimensions of
    the interval (difference between the preceeding and succsesive midpoint).

    :param directions_radians: array of radian directions
    :return: array of radian intervals
    """

    # Calculate the forward difference appending the first entry to the back
    # of the array. Use modular trickery to ensure the angle is in [-pi,pi]
    forward_diff = (numpy.diff(directions_radians,
                               append=directions_radians[0]) + numpy.pi) % (
                           2 * numpy.pi) - numpy.pi

    # Calculate the backward difference prepending the last entry to the front
    # of the array. Use modular trickery to ensure the angle is in [-pi,pi]
    backward_diff = (numpy.diff(directions_radians,
                                prepend=directions_radians[-1]) + numpy.pi) % (
                            2 * numpy.pi) - numpy.pi

    # The interval we are interested in is the average of the forward and backward
    # differences.
    return (forward_diff + backward_diff) / 2


def get_constraint_matrix(directions_radians: numpy.ndarray) -> numpy.ndarray:
    """
    Define the matrix M that can be used in the matrix product M@D (with D the
    directional distribution) such that:

            M@D = [1,a1,b1,a2,b2]^T

    with a1,b1 etc the directional moments at a given frequency.

    :param directions_radians: array of radian directions
    :return:
    """
    number_of_dir = len(directions_radians)
    constraints = numpy.zeros((5, number_of_dir))
    direction_increment = get_direction_increment(directions_radians)
    constraints[0, :] = direction_increment
    constraints[1, :] = direction_increment * numpy.cos(directions_radians)
    constraints[2, :] = direction_increment * numpy.sin(directions_radians)
    constraints[3, :] = direction_increment * numpy.cos(2 * directions_radians)
    constraints[4, :] = direction_increment * numpy.sin(2 * directions_radians)
    return constraints


def get_rhs(a1: numpy.ndarray, b1: numpy.ndarray, a2: numpy.ndarray,
            b2: numpy.ndarray) -> numpy.ndarray:
    """
    Define the matrix rhs that for each row contains the directional moments
    at a given frequency:

    rhs = [ 1, a1[0],b1[0],a2[0],b2[0],
            |    |    |      |    |
            N, a1[0],b1[0],a2[0],b2[0] ]

    These rows are use as the "right hand side" in the linear constraints
    (see get_constraint_matrix)

    :param a1: 1d array of cosine directional moment as function of frequency,
    length [number_of_frequencies]

    :param b1: 1d array of sine directional moment as function of frequency,
    length [number_of_frequencies]

    :param a2: 1d array of double angle cosine directional moment as function
    of frequency, length [number_of_frequencies]

    :param b2: 1d array of double angle sine directional moment as function of
    frequency, length [number_of_frequencies]

    :return: array ( number of frequencies by 5) that for each row contains
    the directional moments at a given frequency
    """
    rhs = numpy.array([numpy.ones_like(a1), a1, b1, a2, b2]).transpose()
    return rhs



"""
Contents: Simple local caching of spectral requests.

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Classes:
- `WaveFleetResource`, remote resource specification for grabbing data from
   wavefleet

Functions:



Use:

"""

# Import
# =============================================================================


from roguewave.filecache.remote_resources import RemoteResource
from roguewave import filecache
from botocore.exceptions import ClientError
from roguewave.filecache.exceptions import _RemoteResourceUriNotFound
from boto3 import resource
from typing import Sequence,List, Iterable
# Constants
# =============================================================================


# Spotter Cache setup.
CACHE_NAME = 'spectral_cache'
CACHE_PATH = '~/temporary_roguewave_files/spectral_cache'
CACHE_SIZE_GB = 5


# Classes
# =============================================================================


class RestartFileResource(RemoteResource):
    """
    Remote resource for downloading data from restart file data given a "URI".
    """
    URI_PREFIX = 'restart-file-s3://'

    def __init__(self):
        self.s3 = resource('s3')

    def download(self):
        def _download_file_from_aws(uri: str, filepath: str) -> bool:
            """
            Worker function to download files from s3. Raise error if the
            object does not exist on s3.
            :param uri: valid uri for resource
            :param filepath: valid filepath to download remote object to.
            :return: True on success
            """
            s3 = self.s3
            decoded_uri = _decode_restart_file_cache_uri(uri, self.URI_PREFIX)
            obj = s3.Object(decoded_uri['bucket'], decoded_uri['key'])

            try:
                if decoded_uri["start"] == 0 and decoded_uri["stop"] == -1:
                    response = obj.download_file(filepath)
                else:
                    data = obj.get(Range=f'bytes={decoded_uri["start"]}-'
                                     f'{decoded_uri["stop"]-1}')['Body']
                    with open(filepath,'wb') as file:
                        file.write(data.read())

            except ClientError as e:
                raise _RemoteResourceUriNotFound(
                    f'Error downloading from {uri}. \n'
                    f'Error code: {e.response["Error"]["Code"]} '
                    f'Error message: {e.response["Error"]["Message"]}'
                )
            return True

        return _download_file_from_aws


# Functions
# =============================================================================
def get_data(uri:Sequence, start_byte:Sequence, stop_byte:Sequence,
                 cache_name: str = None
             ) -> List[bytearray]:
    """

    :return:
    """
    if cache_name is None:
        cache_name = CACHE_NAME

    if not _exists():
        _create_cache(cache_name)

    uris = [_restart_file_cache_uri(x, y, z) for
            x,y,z in zip(uri,start_byte,stop_byte)]

    filepaths = filecache.filepaths(uris, cache_name=cache_name)[0]
    if isinstance(filepaths,str):
        filepaths = [filepaths]

    output = []
    for filepath in filepaths:
        with open( filepath,'rb') as file:
            output.append(bytearray(file.read()))
    return output


# Module internal functions
# =============================================================================
def _restart_file_cache_uri(uri:str, start_byte:int, stop_byte:int) -> str:
    return f'restart-file-{uri}:{start_byte},{stop_byte}'

def _decode_restart_file_cache_uri(uri:str, uri_prefix) -> dict:
    bucket, key_byterange = uri.replace(uri_prefix, '').split('/', maxsplit=1)
    key, byterange = key_byterange.split(':')
    start, stop = byterange.split(',')
    return {'bucket':bucket, 'key':key, 'start':int(start),'stop':int(stop)}

def _create_cache(cache_name):
    if not _exists():
        restartfileresource = RestartFileResource()
        filecache.create_cache(cache_name,
                               cache_path=CACHE_PATH,
                               resources=[restartfileresource],
                               cache_size_GB=CACHE_SIZE_GB)


def _exists() -> bool:
    """
    Check if a spotter cache already exists
    :return: True or False
    """
    return filecache.exists(CACHE_NAME)


import numpy
import typing

PHILLIPS_CONSTANT = 0.0081
GRAVITATIONAL_CONSTANT = 9.81

def pierson_moskowitz_frequency(frequency,
                                peak_frequency,
                                alpha=PHILLIPS_CONSTANT,
                                g=GRAVITATIONAL_CONSTANT):
    """
    Pierson Moskowitz variance-density spectrum with frequency in Hz as
    dependant variable. See e.g. Holthuijsen "Waves in Oceanic Water."

    :param frequency: frequency in Hz (scalar or array)
    :param peak_frequency: peak frequency in Hz
    :param alpha: Phillips constant (default 0.0081)
    :param g: gravitational acceleration (default 9.81)
    :return:
    """
    return alpha * g ** 2 * (2 * numpy.pi) ** -4 * frequency ** -5 * numpy.exp(
        - 5 / 4 * (peak_frequency / frequency) ** 4)


def pierson_moskowitz_angular_frequency(
        angular_frequency: typing.Union[numpy.ndarray, float],
        peak_angular_frequency: float, alpha=PHILLIPS_CONSTANT, g=GRAVITATIONAL_CONSTANT):
    """
    Pierson Moskowitz variance-density spectrum with angular frequency (rad/s) as
    dependant variable. See e.g. Holthuijsen "Waves in Oceanic Water."

    :param frequency: frequency in rad/s (scalar or array)
    :param peak_frequency: peak frequency in rad/s
    :param alpha: Phillips constant (default 0.0081)
    :param g: gravitational acceleration (default 9.81)
    :return: 
    """
    return alpha * g ** 2 * angular_frequency ** -5 * numpy.exp(
        - 5 / 4 * (peak_angular_frequency / angular_frequency) ** 4)


def parametric( shape, directional_shape ):
    pass


"""
Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Module defining the restart file class. This is the primary class to interact
    with wavewatch3 restart file data.

Public Classes:
- `RestartFile`, file to interact with restart files from WW3.

Functions:
- N/A

How To Use This Module
======================
(See the individual functions for details.)

1. import
2. create a restart file object.
"""

import numpy
from roguewave.wavewatch3.resources import Resource
from roguewave.wavewatch3.model_definition import Grid, \
    LinearIndexedGridData
from roguewave.wavetheory.lineardispersion import \
    inverse_intrinsic_dispersion_relation, \
    jacobian_wavenumber_to_radial_frequency
from roguewave.wavewatch3.restart_file_metadata import MetaData
from typing import Sequence, Union, Tuple, Dict
from roguewave.interpolate.nd_interp import NdInterpolator
from datetime import datetime
from functools import cache
from roguewave.tools.time import to_datetime_utc, to_datetime64
from roguewave.interpolate.geometry import TrackSet
from xarray import Dataset, concat
from multiprocessing.pool import ThreadPool
from tqdm import tqdm
from roguewave.wavespectra import FrequencyDirectionSpectrum

MAXIMUM_NUMBER_OF_WORKERS = 10

class RestartFile(Sequence):
    _start_record = 2

    def __init__(self,
                 grid:Grid,
                 meta_data:MetaData,
                 resource:Resource,
                 depth:LinearIndexedGridData=None,
                 return_freq_energy_density=True,
                 parallel=True):

        self._grid = grid
        self._meta_data = meta_data
        self.resource = resource
        self._dtype = numpy.dtype("float32").newbyteorder(meta_data.byte_order)
        self._convert = return_freq_energy_density
        self.parallel = parallel

        if depth is None:
            _depth = numpy.inf \
                * numpy.ones((self.number_of_spatial_points,),dtype='float32')
            self._depth = LinearIndexedGridData(_depth,grid)
        else:
            self._depth = depth

    def set_return_freq_energy_density(self):
        self._convert = True

    def set_return_k_action_density(self):
        self._convert = False

    @property
    def frequency(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of frequencies
        """
        return self._grid.frequencies

    @property
    def direction(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of directions
        """
        return self._grid.directions
    
    @property
    def latitude(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of latitudes.
        """
        return self._grid.latitude

    @property
    def longitude(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of longitudes.
        """
        return self._grid.longitude

    def coordinates(self, index:Union[slice,int,numpy.ndarray]
                    )->Tuple[Union[float,numpy.ndarray ],
                             Union[float,numpy.ndarray ]]:
        """
        Return the latitude and longitude as a function of the linear index.
        :param index: linear index
        :return:  ( latitude(s), longitude(s)
        """
        ilon = self._grid.longitude_index(index)
        ilat = self._grid.latitude_index(index)
        return self.latitude[ilat], self.longitude[ilon]

    @property
    def number_of_directions(self) -> int:
        """
        :return: number of directions
        """
        return len(self.direction)

    @property
    def number_of_frequencies(self) -> int:
        """
        :return: number of frequencies
        """
        return len(self.frequency)
    
    @property
    def number_of_latitudes(self) -> int:
        """
        :return: number of latitudes.
        """
        return len(self.latitude)

    @property
    def number_of_longitudes(self) -> int:
        """
        :return: number of longitudes.
        """
        return len(self.longitude)

    @property
    def number_of_spatial_points(self) -> int:
        """
        :return: Number of spatial points in the restart file. This only counts
            the number of sea points and is *not* equal to
            self.number_of_latitudes * self.number_of_longitudes
            Also referred to as "NSEA" in wavewatch III.
        """
        return self._grid.number_of_spatial_points

    @property
    def number_of_spectral_points(self) -> int:
        """
        :return: number of spectral points.
        """
        return self.number_of_frequencies * self.number_of_directions

    @property
    def time(self) -> datetime:
        """
        :return: Valid time of the restart file.
        """
        return self._meta_data.time

    def __len__(self) -> int:
        """
        :return: We consider a restart file as a container of spectra, so its
            length is the number of spatial points.
        """
        return self.number_of_spatial_points

    def _sliced_index(self,s:slice) -> numpy.ndarray:
        """
        Get wavenumber action density spectra at sliced indices.
        :param s: slice
        :return: wavenumber spectra from sliced indices. Returned data is of
        type float32 and has the shape:
                (s.stop - s.start,
                number_of_frequencies,
                number_of_directions)
        """
        if isinstance(s, (int,numpy.int32,numpy.int64)):
            s = slice(s, s + 1, 1)

        elif not isinstance( s, slice ):
            raise ValueError(f'Cannot use type {type(s)} as a spatial index.'
                             f' Use a slice or int instead.')

        start,stop,step = s.indices(self.number_of_spatial_points)
        start = start + self._start_record
        stop = stop + self._start_record
        if step != 1:
            # We cannot use contiguous IO in this case, recast as a fancy index
            indices = numpy.arange( start, stop, step=step )
            return self._fancy_index(indices)

        byte_slice = slice( self._byte_index(start),
                            self._byte_index(stop),1
                            )

        # Read raw data, cast as numpy array,
        number_of_spatial_points = stop-start
        if number_of_spatial_points == self.number_of_spatial_points:
            # If we read all data we just call the read function. This makes
            # little difference for a local file- but for an aws object this
            # allows us to use the much more efficient get_object.
            self.resource.seek( self._byte_index(self._start_record) )
            data = self.resource.read()
        else:
            data = self.resource.read_range(byte_slice)[0]
        data = numpy.frombuffer(data, dtype=self._dtype,
            count=self.number_of_spectral_points*number_of_spatial_points)

        return numpy.reshape(
            data , (
                stop - start,
                self.number_of_frequencies,
                self.number_of_directions
            )
        )

    def __getitem__(self, s:Union[slice,numpy.ndarray,Sequence,int]
                    ) -> Dataset:
        """
        Dunder method for the Sequence protocol. If obj is an instance of
        RestartFile, this allows us to use `obj[10:13]` to get spectra with
        linear indices 10 to 12.

        Input can either be a integer, slice, or fancy indexing through
        a sequency of numbers or a 1d numpy integer array.

        :param s: slice, integer, 1d numpy integer array, Sequence of integers.
        :return: We return the requested spectra. Depending on how we setup
            the object we return frequency energy density spectra (default) or
            wavenumber spectra. Returned data is of type float32 and has
            the shape:
                (number_of_spatial_points_requested,
                number_of_frequencies,
                number_of_directions)
        """
        if isinstance(s, Sequence) or isinstance(s, numpy.ndarray):
            data = self._fancy_index(s)
        else:
            data = self._sliced_index(s)

        if self._convert:
            # Are we using raw wavenumbers or frequency spectra
            jacobian = self.to_frequency_energy_density(s)
            data = data * jacobian

        if isinstance(s,slice):
            s = list(range(*s.indices(self.number_of_spatial_points)))
        elif isinstance(s,(int,numpy.int32,numpy.int64)):
            s = [s]

        coords = self.coordinates(s)
        return Dataset(
            data_vars={
                "variance_density": (
                    ('point_index','frequency','direction'),data ),
                "longitude": (
                    ('point_index'),coords[1] ),
                "latitude": (
                    ('point_index'), coords[0]),
                "linear_index": (
                    ('point_index'), s),
            },
            coords={
                "point index": numpy.arange(0,len(s) ),
                'frequency':self.frequency,
                'direction':self.direction
            })


    def _fancy_index(self, indices:Union[Sequence,numpy.ndarray]
                     ) -> numpy.ndarray:
        """
        Get wavenumber action density spectra at given indices.
        :param indices: linear indices we want spectra from (Sequence or numpy
            array of integers).
        :return: wavenumber spectra from sliced indices. Returned data is of
        type float32 and has the shape:
                (len(indices),
                number_of_frequencies,
                number_of_directions)
        """
        if isinstance(indices, Sequence):
            indices = numpy.array(indices,dtype='int32')

        indices = indices + self._start_record
        slices = [ slice(self._byte_index(index),self._byte_index(index+1),1)
                   for index in indices ]

        data_for_slices = self.resource.read_range(slices)
        out = []
        for data in data_for_slices:
            data = numpy.frombuffer(data, dtype=self._dtype)
            out.append(
                numpy.reshape(
                    data , (
                        self.number_of_frequencies,
                        self.number_of_directions
                    )
                )
            )
        return numpy.array(out)

    def _byte_index(self,index) -> int:
        """
        Get a byte index of a requested record. Record sizes are always
        number_of_spectral_points * 4 bytes. Note that the first
        self._start_record's contain header information. Spectral data starts
        at index = self._start_record
        :param index: Linear index
        :return: Linear byte index
        """
        return index * self._meta_data.record_size_bytes

    def interpolate_in_space(self,
                             latitude:Union[numpy.ndarray,float],
                             longitude:Union[numpy.ndarray,float]) -> Dataset:
        """
        Extract interpolated spectra at given latitudes and longitudes.
        Input can be either a single latitude and longitude pair, or a
        numpy array of latitudes and longitudes.

        :param latitude: latitudes to get interpolated spectra
        :param longitude: longitudes to get interpolated spectra
        :return: Interpolated spectra. Returned data is of  type float32 and
        has the shape:
                (len(indices),
                number_of_frequencies,
                number_of_directions)
        """
        latitude = numpy.atleast_1d(latitude)
        longitude = numpy.atleast_1d(longitude)
        points = {"latitude":numpy.atleast_1d(latitude),
                  "longitude":numpy.atleast_1d(longitude)}

        periodic_coordinates = {"longitude":360}

        def _get_data(  indices, _dummy ):
            index = self._grid.index(latitude_index=indices[0],
                                     longitude_index=indices[1])

            output = numpy.zeros( (len(index),
                                     self.number_of_frequencies,
                                     self.number_of_directions ) )
            mask = index >= 0
            output[ mask,:,:] = \
                self.__getitem__(index[mask])['variance_density'].values
            output[~mask,:,:] =numpy.nan
            return output


        data_shape = [ len(self.latitude), len(self.longitude),
                       self.number_of_frequencies,self.number_of_directions]
        data_coordinates = (
            ('latitude', self.latitude),
            ('longitude', self.longitude),
            ('frequency',self.frequency),
            ('direction', self.direction)
        )

        interpolator = NdInterpolator(
            get_data=_get_data,
            data_coordinates=data_coordinates,
            data_shape=data_shape,
            interp_coord_names=list(points.keys()),
            interp_index_coord_name='latitude',
            data_periodic_coordinates=periodic_coordinates,
            data_period=None,
            data_discont=None
        )
        return self._create_dataset( interpolator.interpolate(points),
                                     numpy.arange(len(latitude)) )

    def to_wavenumber_action_density(
            self, s:Union[slice, numpy.ndarray, Sequence] ) -> numpy.array:
        """
        Factor that when multiplied with frequency energy density spectra at
        the givem indices converts them to action wavenumber spectra. E.g:
            Ek = to_wavenumber_action_density(slice[10,20)) * Ef[10:20]

        :param s: slice for linear indices.
        :return: Multiplication factor as numpy ndarray of size.
                (number_of_elements_in_slice,number_of_frequencies,1)

        To note, we add the trailing 1 to make it easy to use broadcasting (
            the Jacobian is constant in directional space).
        """
        return 1 / self.to_frequency_energy_density(s)[:,:,None]

    def to_frequency_energy_density(
            self, s:Union[slice, numpy.ndarray, Sequence]
                                    ) -> numpy.ndarray:

        """
        Factor that when multiplied with wavenumber action density spectra at
        the givem indices converts them to frequency energy spectra. E.g:
            Ef = to_frequency_energy_density(slice[10,20)) * Ek[10:20]

        :param s: slice for linear indices.
        :return: Multiplication factor as numpy ndarray of size.
                (number_of_elements_in_slice,number_of_frequencies,1)

        To note, we add the trailing 1 to make it easy to use broadcasting (
            the Jacobian is constant in directional space).

        To convert a wavenumber Action density as a function of radial
        frequency (as stored by ww3) to a frequency Energy density we need to
        multiply the action density with the angular frequency (Action -> to
        Energy) and the proper Jacobians of the transformations. Specficically
                 #
        # 1) transformation from k-> omega
            (jacobian_wavenumber_to_radial_frequency() )
        # 2) transformation from omega -> f    ( 2 * pi )
        # 3) transformation from radians -> degrees   ( pi / 180.)
        """

        # get the depth and make sure it has dimension [ spatial_points , 1 ]
        depth = self._depth[s, None]

        # Get omega and make sure it has dimensions [ 1 , frequency_points ]
        w = self.frequency[None, :] * numpy.pi * 2

        # Repeat w along the spatial axis
        w = numpy.repeat(w, len(depth), axis=0)

        # get the wavenumber
        k = inverse_intrinsic_dispersion_relation(w, depth)

        # Calculate the various (Jacobian) factors
        jac_k_to_w = jacobian_wavenumber_to_radial_frequency(k, depth)
        jac_omega_f = 2*numpy.pi
        jac_rad_to_deg = numpy.pi/180
        action_to_energy = w

        # Return the result
        factor = action_to_energy * jac_rad_to_deg * jac_omega_f * jac_k_to_w
        return factor[:,:,None]

    @cache
    def header_bytes(self) -> bytes:
        """
        Return the header bytes, the first self._record_start records. This is
        primarily used to created a new restart file that has the same metadata
        as the current restart file.

        We cache this information to allow for more rapid retrieval if the
        information is stored on s3.

        :return: raw bytes of the header.
        """
        s = slice(self._byte_index(0),
                  self._byte_index(self._start_record),1)
        return self.resource.read_range(s)[0]

    @cache
    def tail_bytes(self) -> bytes:
        """
        After all spectral data in the restart file there is a bunch of
        additional information regarding currents (needed for uniqueness of
        energy to action) and other information. We do not currently parse this
        information in any way. Instead, if we want to create an updated
        version of the restart file (e.g. for data assimilation) we merely
        append the information from the source (or background) file. This
        function is a helper method to retrieve that information.

        We cache this information to allow for more rapid retrieval if the
        information is stored on s3.

        :return: raw bytes of the tail.
        """
        self.resource.seek(self._byte_index(self._start_record
                                            + self.number_of_spatial_points))
        return self.resource.read()

    def variance(self, latitude_slice:slice, longitude_slice:slice
                 ) -> numpy.ndarray:
        """
        Calculate the variance at sliced indices for latitudes and longitudes.
        Returns a 2d array with constant latitudes along rows and
        constant longitudes along colunns

        :param latitude_slice: latitude index range as slice
        :param longitude_slice: longitude index range as slice
        :return:
        """
        index = self._grid.index(
            latitude_index=latitude_slice,
            longitude_index=longitude_slice,
            valid_only=True)

        linear_indexed_variance = self.variance_linear_index(index)
        return self._grid.project( lon_slice=longitude_slice,
                                   lat_slice=latitude_slice,
                                   var=linear_indexed_variance)

    def variance_linear_index(self, index) -> numpy.ndarray:
        """
        Calculate the variance at linear indices. Returns a 1d of variances
        at requested indices

        :param latitude_slice: latitude index range as slice
        :param longitude_slice: longitude index range as slice
        :return:
        """
        toggle = not self._convert
        if toggle:
            self.set_return_freq_energy_density()
        spectra = self[index]['variance_density'].values
        if toggle:
            self.set_return_k_action_density()

        delta_f = self._grid.frequency_step()[None,:,None]
        delta_dir = self._grid.direction_step()[None,None,:]
        return numpy.sum( delta_f*delta_dir*spectra,axis=(1,2) )


    def number_of_header_bytes(self) -> int:
        """
        :return: length of the header in bytes.
        """
        return len(self.header_bytes())

    def number_of_tail_bytes(self) -> int:
        """
        :return: length of the tail in bytes.
        """
        return len(self.tail_bytes())

    def size_in_bytes(self):
        """
        :return: Total size in bytes of a restart file.
        """
        return self.number_of_tail_bytes() + self.number_of_header_bytes() + \
               self.number_of_spatial_points * \
               self.number_of_spectral_points * self._dtype.itemsize

    def linear_indices(self,indices:Union[slice,numpy.ndarray,Sequence]):
        return numpy.arange(self.number_of_spatial_points)[indices]

    def _create_dataset(self, variance_density,
                        linear_indices: Union[slice,numpy.ndarray,Sequence]):

        coords = self.coordinates(linear_indices)
        return Dataset(
            data_vars={
                "variance_density": (
                    ('point_index','frequency','direction'),variance_density),
                "longitude": (
                    ('point_index'),coords[1] ),
                "latitude": (
                    ('point_index'),coords[0]),
                "linear_index": (
                    ('point_index'),linear_indices),
                "time": self.time
            },
            coords={
                "point index": numpy.arange(0,len(linear_indices) ),
                'frequency':self.frequency,
                'direction':self.direction
            })



class RestartFileStack:
    def __init__(self, restart_files: Sequence[RestartFile],parallel = True):
        self._restart_files = restart_files
        self._grid = restart_files[0]._grid
        self._time = to_datetime_utc([ x.time for x in restart_files ])
        self.parallel = parallel

        # To make the progress bar work we need to store progress across
        # calls somewhere.
        self._progres = {"position":0, "total":None, "leave":True}

    @property
    def frequency(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of frequencies
        """
        return self._grid.frequencies

    @property
    def direction(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of directions
        """
        return self._grid.directions

    @property
    def latitude(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of latitudes.
        """
        return self._grid.latitude

    @property
    def longitude(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of longitudes.
        """
        return self._grid.longitude

    def coordinates(self, index: Union[slice, int, numpy.ndarray]
                    ) -> Tuple[Union[float, numpy.ndarray],
                               Union[float, numpy.ndarray]]:
        """
        Return the latitude and longitude as a function of the linear index.
        :param index: linear index
        :return:  ( latitude(s), longitude(s)
        """
        ilon = self._grid.longitude_index(index)
        ilat = self._grid.latitude_index(index)
        return self.latitude[ilat], self.longitude[ilon]

    @property
    def number_of_directions(self) -> int:
        """
        :return: number of directions
        """
        return len(self.direction)

    @property
    def number_of_frequencies(self) -> int:
        """
        :return: number of frequencies
        """
        return len(self.frequency)

    @property
    def number_of_latitudes(self) -> int:
        """
        :return: number of latitudes.
        """
        return len(self.latitude)

    @property
    def number_of_longitudes(self) -> int:
        """
        :return: number of longitudes.
        """
        return len(self.longitude)

    @property
    def number_of_spatial_points(self) -> int:
        """
        :return: Number of spatial points in the restart file. This only counts
            the number of sea points and is *not* equal to
            self.number_of_latitudes * self.number_of_longitudes
            Also referred to as "NSEA" in wavewatch III.
        """
        return self._grid.number_of_spatial_points

    @property
    def number_of_spectral_points(self) -> int:
        """
        :return: number of spectral points.
        """
        return self.number_of_frequencies * self.number_of_directions

    @property
    def time(self) -> Sequence[datetime]:
        return self._time

    def __len__(self):
        return len(self._restart_files)

    def __getitem__(self, nargs) -> FrequencyDirectionSpectrum:
        if len(nargs) == 2:
            time_index, linear_index = nargs
        elif len(nargs) == 3:
            time_index, lat_index, lon_index = nargs
            linear_index = self._grid.index(lat_index,
                                            lon_index,valid_only=True)
        else:
            raise ValueError('unexpected number of indices')

        if isinstance(time_index,(int,numpy.int32,numpy.int64)):
            time_index = [time_index]

        fancy_index = not isinstance(time_index,slice)
        if fancy_index:
            _input = list(zip( time_index,linear_index ))
        else:
            time_index = list(range(*time_index.indices(len(self))))
            _input = [ (it, linear_index) for it in time_index]

        def _worker( arg ):
            return self._restart_files[arg[0]][arg[1]]

        self._init_progress_bar(len(_input))
        if self.parallel and len(_input) > 1:
            with ThreadPool(processes=MAXIMUM_NUMBER_OF_WORKERS) as pool:
                data = list(
                    tqdm(
                        pool.imap(_worker, _input),
                        total=self._progres['total'],
                        initial=self._progres['position'],
                        leave=self._progres['leave']
                    )
                )
        else:
            disable_progress_bar = False
            if len(_input) == 1:
                disable_progress_bar = True

            data = list(
                tqdm(
                    map(_worker, _input),
                    total=self._progres['total'],
                    disable=disable_progress_bar,
                    initial=self._progres['position'],
                    leave = self._progres['leave']
                )
            )
        self._update_progress_bar(len(_input))

        data = concat( data , dim='time')
        data.coords['time'] = to_datetime64([self.time[it] for it in time_index])
        return FrequencyDirectionSpectrum(data)

    def _init_progress_bar(self, total):
        if self._progres['total'] is None:
            self._progres['total'] = total
            self._progres['position'] = 0
            self._progres['leave'] = True
        else:
            if total + self._progres['position'] == self._progres['total']:
                self._progres['leave'] = True
            else:
                self._progres['leave'] = False

    def _update_progress_bar(self,number):
        self._progres['position'] += number

        if self._progres['position'] == self._progres['total']:
            self._progres['total'] = None
            self._progres['position'] = 0

    def interpolate(self,latitude:Union[numpy.ndarray,float],
                        longitude:Union[numpy.ndarray,float],
                        time:Union[numpy.ndarray,numpy.datetime64,datetime]
                    ) -> FrequencyDirectionSpectrum:
        """
        Extract interpolated spectra at given latitudes and longitudes.
        Input can be either a single latitude and longitude pair, or a
        numpy array of latitudes and longitudes.

        :param latitude: latitudes to get interpolated spectra
        :param longitude: longitudes to get interpolated spectra
        :return: Interpolated spectra. Returned data is of  type float32 and
        has the shape:
                (len(indices),
                number_of_frequencies,
                number_of_directions)
        """
        time = to_datetime64(time)
        points = {"time":numpy.atleast_1d(time),
                  "latitude":numpy.atleast_1d(latitude),
                  "longitude":numpy.atleast_1d(longitude),
                  }

        periodic_coordinates = {"longitude":360}

        def _get_data(  indices , idims):
            time_index = indices[0]
            index = self._grid.index(latitude_index=indices[1],
                                     longitude_index=indices[2])

            output = numpy.zeros( (
                len(index),
                self.number_of_frequencies,
                self.number_of_directions ) )
            mask = index >= 0

            output[  mask,:,:] = numpy.squeeze(
                self.__getitem__(
                    (time_index[mask],index[mask]))['variance_density']
            )
            output[ ~mask,:,:] =numpy.nan
            return output

        data_shape = [ len(self.time), len(self.latitude), len(self.longitude),
                       self.number_of_frequencies,self.number_of_directions]

        data_coordinates = (
            ('time',to_datetime64(self.time)),
            ('latitude', self.latitude),
            ('longitude', self.longitude),
            ('frequency',self.frequency),
            ('direction', self.direction)
        )

        interpolator = NdInterpolator(
            get_data=_get_data,
            data_coordinates=data_coordinates,
            data_shape=data_shape,
            interp_coord_names=list(points.keys()),
            interp_index_coord_name='time',
            data_periodic_coordinates=periodic_coordinates,
            data_period=None,
            data_discont=None
        )

        self._init_progress_bar(len(latitude)*8)
        dataset = interpolator.interpolate(points)

        return FrequencyDirectionSpectrum(Dataset(
            data_vars={
                "variance_density": (
                    ('time','frequency','direction'),dataset ),
                "longitude": (
                    ('time'),longitude ),
                "latitude": (
                    ('time'),latitude),
            },
            coords={
                'time':time,
                'frequency':self.frequency,
                'direction':self.direction
            }))

    def interpolate_tracks(self,tracks:TrackSet
                          ) -> Dict[str, FrequencyDirectionSpectrum]:
        return {
            _id:self.interpolate(x.latitude,x.longitude,x.time)
                                             for _id,x in tracks.tracks.items()
        }




"""
Contents: Spectral estimators

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Spectral estimators that can be used to create a 2D spectrum from buoy
observations

Classes:

- `None

Functions:

- `mem`, maximum entrophy method
- `mem2`, ...

How To Use This Module
======================
(See the individual functions for details.)
"""
from roguewave.wavespectra.estimators.mem2 import mem2
from roguewave.wavespectra.estimators.mem import mem
from roguewave.wavespectra.estimators.loglikelyhood import log_likelyhood
import numpy

# -----------------------------------------------------------------------------
#                       Boilerplate Interfaces
# -----------------------------------------------------------------------------
def spec2d_from_spec1d(
        e:numpy.ndarray,
        a1:numpy.ndarray,
        b1:numpy.ndarray,
        a2:numpy.ndarray,
        b2:numpy.ndarray,
        direction:numpy.ndarray,
        method: str = 'mem2') -> numpy.ndarray:
    """
    Construct a 2D spectrum based on the 1.5D spectrum and a spectral
    reconstruction method.

    :param number_of_directions: length of the directional vector for the
    2D spectrum. Directions returned are in degrees

    :param method: Choose a method in ['mem','ridge','msem']
        mem: maximum entrophy (in the Boltzmann sense) method
        Lygre, A., & Krogstad, H. E. (1986). Explicit expression and
        fast but tends to create narrow spectra anderroneous secondary peaks.

        mem2: use entrophy (in the Shannon sense) to maximize. Likely
        best method see- Benoit, M. (1993).

        log: solve underdetermined constrained problem using ridge regression
        as regulizer. Performant because efficient solutions to constrained
        quadratic programming problems exist. The spectra compare well to
        nlmen- though are somewhat broader in general. Does not suffer from
        the spurious peak issue of mem. Default method for there reasons.

    REFERENCES:
    Benoit, M. (1993). Practical comparative performance survey of methods
        used for estimating directional wave spectra from heave-pitch-roll data.
        In Coastal Engineering 1992 (pp. 62-75).

    Lygre, A., & Krogstad, H. E. (1986). Maximum entropy estimation of the
        directional distribution in ocean wave spectra.
        Journal of Physical Oceanography, 16(12), 2052-2060.

    """


    # Jacobian to transform distribution as function of radian angles into
    # degrees.
    Jacobian = numpy.pi / 180

    if method.lower() in ['maximum_entropy_method', 'mem']:
        # reconstruct the directional distribution using the maximum entropy
        # method.
        directional_distribution = mem(direction * numpy.pi / 180, a1, b1, a2,
                                       b2) * Jacobian
    elif method.lower() in ['log_likelyhood', 'log']:
        directional_distribution = log_likelyhood(direction * numpy.pi / 180,
                                                  a1, b1, a2,
                                                  b2) * Jacobian
    elif method.lower() in ['maximum_entrophy_method2', 'mem2']:
        directional_distribution = mem2(
            direction * numpy.pi / 180, a1, b1, a2,
            b2) * Jacobian
    else:
        raise Exception(f'unsupported spectral estimator method: {method}')

    # We return a 2D wave spectrum object.
    return e[:, None] * directional_distribution



from roguewave.wavewatch3.restart_file import RestartFile, RestartFileStack
from roguewave.wavewatch3.io import clone_restart_file, open_restart_file, \
    write_partial_restart_file, write_restart_file, \
    reassemble_restart_file_from_parts, \
    open_restart_file_stack

from roguewave.wavewatch3.resources import create_resource
from roguewave.wavewatch3.restart_file_metadata import read_header, \
    MetaData
from roguewave.wavewatch3.model_definition import read_model_definition, \
    Grid


import numpy
from roguewave.interpolate.dataset import (
    interpolate_dataset_grid,
    interpolate_dataset_along_axis
)
from roguewave.tools.math import wrapped_difference
from roguewave.tools.time import to_datetime64
from roguewave.wavetheory.lineardispersion import \
    inverse_intrinsic_dispersion_relation
from roguewave.wavespectra.estimators import mem2
from roguewave.wavespectra.parametric import pierson_moskowitz_frequency
from typing import Iterator, Hashable, Iterable, TypeVar
from xarray import Dataset, DataArray, concat
from xarray.core.coordinates import DatasetCoordinates

_NAME_F = 'frequency'
_NAME_D = 'direction'
_NAME_T = 'time'
_NAME_E = 'variance_density'
_NAME_a1 = 'a1'
_NAME_b1 = 'b1'
_NAME_a2 = 'a2'
_NAME_b2 = 'b2'
_NAME_LAT = 'latitude'
_NAME_LON = 'longitude'
_NAME_DEPTH = 'depth'
_NAMES_2D = (_NAME_F, _NAME_D, _NAME_T, _NAME_E, _NAME_LAT, _NAME_LON)
_NAMES_1D = (_NAME_F, _NAME_T, _NAME_E, _NAME_LAT, _NAME_LON,
             _NAME_a1, _NAME_b1, _NAME_a2, _NAME_b2)

_T = TypeVar("_T")

class DatasetWrapper():
    """
    A class that wraps a dataset object and passes through some of its primary
    functionality (get/set etc.). Used here mostly to make explicit what parts
    of the Dataset interface we actually expose in frequency objects. Note that
    we do not claim- or try to obtain completeness here. If full capabilities
    of the dataset object are needed we can simple operate directly on the
    dataset object itself.
    """
    def __init__(self, dataset: Dataset):
        self.dataset = dataset

    def __getitem__(self, item):
        return self.dataset.__getitem__(item)

    def __setitem__(self, key, value) -> None:
        return self.dataset.__setitem__(key, value)

    def __copy__(self:_T) -> _T:
        cls = self.__class__
        return cls(self.dataset.copy())

    def __len__(self):
        return len(self.dataset)

    def copy(self:_T, deep=True) -> _T:
        if deep:
            return self.__deepcopy__({})
        else:
            return self.__copy__()

    def __deepcopy__(self:_T, memodict) -> _T:
        cls = self.__class__
        return cls(self.dataset.copy(deep=True))

    def coords(self) -> DatasetCoordinates:
        return self.dataset.coords

    def keys(self):
        return self.dataset.keys()

    def __contains__(self, key: object) -> bool:
        return key in self.dataset

    def __iter__(self) -> Iterator[Hashable]:
        return self.dataset.__iter__()

    @classmethod
    def concat_from_list(cls: _T, _list: Iterable[_T], dim='time') -> _T:
        """
        Classmethod to aggregate a list of wavespectra and concatenate them
        into a single spectral object. For instance, from spotter observations
        we may get 1 spectrum per timestamp- and we want to aggregate all
        spectra in a single object.

        :param _list: list (Iterable) of WaveSpectra
        :param dim: the dimension over which to concatenate the spectra,
            usually time.
        :return: Wavespectrum object.
        """
        return cls(concat( [x.dataset for x in _list], dim=dim))


class WaveSpectrum(DatasetWrapper):
    frequency_units = 'Hertz'
    angular_units = 'Degrees'
    spectral_density_units = 'm**2/Hertz'
    angular_convention = 'Wave travel direction (going-to), ' \
                         'measured anti-clockwise from East'
    bulk_properties = (
        'm0', 'hm0', 'tm01', 'tm02', 'peak_period', 'peak_direction',
        'peak_directional_spread', 'mean_direction', 'mean_directional_spread',
        'peak_frequency',
        'peak_wavenumber', 'latitude', 'longitude', 'time',
        'is_sea_spectrum', 'is_swell_spectrum')

    def __init__(self, dataset: Dataset):
        super(WaveSpectrum, self).__init__(dataset)

    def __add__(self:_T, other: _T) -> _T:
        spectrum = self.copy(deep=True)
        spectrum[_NAME_E] = (spectrum[_NAME_E] +
                             other[_NAME_E])
        return spectrum

    def __sub__(self: _T, other: _T) -> _T:
        spectrum = self.copy(deep=True)
        spectrum[_NAME_E] = (spectrum[_NAME_E] -
                             other[_NAME_E])
        return spectrum

    def __neg__(self:_T) -> _T:
        """
        Negate self- that is -spectrum
        :return: spectrum with all spectral values taken to have the opposite
            sign.
        """
        spectrum = self.copy(deep=True)
        spectrum[_NAME_E] = -spectrum[_NAME_E]
        return spectrum

    def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf
                         ) -> DataArray:
        """
        Calculate a "frequency moment" over the given range. A frequency moment
        here refers to the integral:

                    Integral-over-frequency-range[ e(f) * f**power ]

        :param power: power of the frequency
        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: frequency moment
        """
        _range = self._range(fmin, fmax)
        return (
            self.e.isel({_NAME_F: _range}) * self.frequency[_range] ** power
        ).integrate(coord=_NAME_F)

    @property
    def number_of_frequencies(self) -> int:
        """
        :return: number of frequencies
        """
        return len(self.frequency)

    @property
    def spectral_values(self) -> DataArray:
        """
        :return: Spectral levels
        """
        return self.dataset[_NAME_E]

    @property
    def radian_frequency(self) -> DataArray:
        """
        :return: Radian frequency
        """
        data_array = self[_NAME_F] * 2 * numpy.pi
        data_array.name = 'radian_frequency'
        return data_array

    @property
    def latitude(self) -> DataArray:
        """
        :return: latitudes
        """
        return self[_NAME_LAT]

    @property
    def longitude(self) -> DataArray:
        """
        :return: longitudes
        """
        return self[_NAME_LON]

    @property
    def time(self) -> DataArray:
        """
        :return: Time
        """
        return self[_NAME_T]

    @property
    def e(self) -> DataArray:
        """
        :return: 1D spectral values (directionally integrated spectrum).
            Equivalent to self.spectral_values if this is a 1D spectrum.
        """
        return self[_NAME_E]

    @property
    def a1(self) -> DataArray:
        """
        :return: normalized Fourier moment cos(theta)
        """
        return self[_NAME_a1]

    @property
    def b1(self) -> DataArray:
        """
        :return: normalized Fourier moment sin(theta)
        """
        return self[_NAME_b1]

    @property
    def a2(self) -> DataArray:
        """
        :return: normalized Fourier moment cos(2*theta)
        """
        return self[_NAME_a2]

    @property
    def b2(self) -> DataArray:
        """
        :return: normalized Fourier moment sin(2*theta)
        """
        return self[_NAME_b2]

    @property
    def A1(self) -> DataArray:
        """
        :return: Fourier moment cos(theta)
        """
        return self.a1 * self.e

    @property
    def B1(self) -> DataArray:
        """
        :return: Fourier moment sin(theta)
        """
        return self.b1 * self.e

    @property
    def A2(self) -> DataArray:
        """
        :return: Fourier moment cos(2*theta)
        """
        return self.a2 * self.e

    @property
    def B2(self) -> DataArray:
        """
        :return: Fourier moment sin(2*theta)
        """
        return self.b2 * self.e

    @property
    def frequency(self) -> DataArray:
        """
        :return: Frequencies (Hz)
        """
        return self[_NAME_F]

    def m0(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Zero order frequency moment. Also referred to as variance or energy.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: variance/energy
        """
        return self.frequency_moment(0, fmin, fmax)

    def m1(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        First order frequency moment. Primarily used in calculating a mean
        period measure (Tm01)

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: first order frequency moment.
        """
        return self.frequency_moment(1, fmin, fmax)

    def m2(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Second order frequency moment. Primarily used in calculating the zero
        crossing period (Tm02)

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: Second order frequency moment.
        """
        return self.frequency_moment(2, fmin, fmax)

    def hm0(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Significant wave height estimated from the spectrum, i.e. waveheight
        h estimated from variance m0. Common notation in literature.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: Significant wave height
        """
        return 4 * numpy.sqrt(self.m0(fmin, fmax))

    def tm01(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Mean period, estimated as the inverse of the center of mass of the
        spectral curve under the 1d spectrum.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: Mean period
        """
        return self.m0(fmin, fmax) / self.m1(fmin, fmax)

    def tm02(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Zero crossing period based on Rice's spectral estimate.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: Zero crossing period
        """
        return numpy.sqrt(self.m0(fmin, fmax) / self.m2(fmin, fmax))

    def peak_index(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Index of the peak frequency of the 1d spectrum within the given range
        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: peak indices
        """
        return self.e.where(self._range(fmin, fmax), 0).argmax(dim=_NAME_F)

    def peak_frequency(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Peak frequency of the spectrum, i.e. frequency at which the spectrum
        obtains its maximum.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: peak frequency
        """
        return self[_NAME_F][self.peak_index(fmin, fmax)]

    def peak_period(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Peak period of the spectrum, i.e. period at which the spectrum
        obtains its maximum.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: peak period
        """
        return 1 / self.peak_frequency(fmin, fmax)

    def peak_direction(self, fmin=0, fmax=numpy.inf) -> DataArray:
        index = self.peak_index(fmin, fmax)
        return self._mean_direction(
            self.a1.isel(**{_NAME_F: index}),
            self.a1.isel(**{_NAME_F: index}))

    def peak_directional_spread(self, fmin=0, fmax=numpy.inf) -> DataArray:
        index = self.peak_index(fmin, fmax)
        a1 = self.a1.isel(**{_NAME_F: index})
        b1 = self.a1.isel(**{_NAME_F: index})
        return self._spread(a1, b1)

    @staticmethod
    def _mean_direction(a1:_T, b1:_T) -> _T:
        return numpy.arctan2(b1, a1) * 180 / numpy.pi

    @staticmethod
    def _spread(a1:_T, b1:_T) -> _T:
        return numpy.sqrt(
            2 - 2 * numpy.sqrt(a1 ** 2 + b1 ** 2)) * 180 / numpy.pi

    @property
    def mean_direction_per_frequency(self) -> DataArray:
        return self._mean_direction(self.a1, self.b1)

    @property
    def mean_spread_per_frequency(self) -> DataArray:
        return self._spread(self.a1, self.b1)

    def _spectral_weighted(self, property: DataArray, fmin=0, fmax=numpy.inf):
        range = {_NAME_F: self._range(fmin, fmax)}

        return numpy.trapz(property.isel(**range) * self.e.isel(**range),
                           self.frequency[range]) / self.m0(fmin, fmax)

    def mean_direction(self, fmin=0, fmax=numpy.inf):
        return self._mean_direction(self.mean_a1(fmin, fmax),
                                    self.mean_b1(fmin, fmax))

    def mean_directional_spread(self, fmin=0, fmax=numpy.inf):
        return self._spread(self.mean_a1(fmin, fmax), self.mean_b1(fmin, fmax))

    def mean_a1(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.a1, fmin, fmax)

    def mean_b1(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.b1, fmin, fmax)

    def mean_a2(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.a2, fmin, fmax)

    def mean_b2(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.b2, fmin, fmax)

    @property
    def depth(self):
        if _NAME_DEPTH in self.dataset:
            return self.dataset[_NAME_DEPTH]
        else:
            return numpy.inf

    @property
    def wavenumber(self) -> DataArray:
        return inverse_intrinsic_dispersion_relation(self.radian_frequency,
                                                     self.depth)

    @property
    def wavelength(self) -> DataArray:
        return 2 * numpy.pi / self.wavenumber

    @property
    def peak_wavenumber(self) -> DataArray:
        index = self.peak_index()
        return inverse_intrinsic_dispersion_relation(
            self.radian_frequency[index], self.depth)

    def bulk_variables(self) -> Dataset:
        dataset = Dataset()
        dataset['significant_waveheight'] = self.significant_waveheight
        dataset['mean_period'] = self.mean_period
        dataset['peak_period'] = self.peak_period()
        dataset['peak_direction'] = self.peak_direction()
        dataset['peak_directional_spread'] = self.peak_directional_spread()
        dataset['mean_direction'] = self.mean_direction()
        dataset['mean_directional_spread'] = self.mean_directional_spread()
        dataset['peak_frequency'] = self.peak_frequency()
        dataset['latitude'] = self.latitude
        dataset['longitude'] = self.longitude
        dataset['timestamp'] = self.time
        return dataset

    @property
    def significant_waveheight(self) -> DataArray:
        return self.hm0()

    @property
    def mean_period(self) -> DataArray:
        return self.tm01()

    @property
    def zero_crossing_period(self) -> DataArray:
        return self.tm02()

    def interpolate(self:_T, coordinates) -> _T:
        return self.__class__(
            interpolate_dataset_grid(coordinates, self.dataset))

    def interpolate_frequency(self:_T, new_frequencies) -> _T:
        return self.__class__(interpolate_dataset_along_axis(
            new_frequencies, self.dataset, coordinate_name='frequency'))

    @property
    def is_sea_spectrum(self) -> DataArray:
        """
        Identify whether or not it is a sea partion. Use 1D method for 2D
        spectra in section 3 of:

        Portilla, J., Ocampo-Torres, F. J., & Monbaliu, J. (2009).
        Spectral partitioning and identification of wind sea and swell.
        Journal of atmospheric and oceanic technology, 26(1), 107-122.

        :return: boolean indicting it is sea
        """
        peak_index = self.peak_index()
        peak_frequency = self.peak_frequency()

        peak_level = \
            self.e.isel(indexers={'frequency': peak_index.values})
        pm_value = pierson_moskowitz_frequency(peak_frequency,
                                               peak_frequency)
        return peak_level == pm_value

    @property
    def is_swell_spectrum(self) -> DataArray:
        return ~self.is_sea_spectrum

    def _range(self, fmin=0.0, fmax=numpy.inf) -> numpy.ndarray:
        return ((self[_NAME_F].values >= fmin) &
                (self[_NAME_F].values < fmax))


class FrequencyDirectionSpectrum(WaveSpectrum):
    def __init__(self, dataset: Dataset):
        super(FrequencyDirectionSpectrum, self).__init__(dataset)
        for name in _NAMES_2D:
            if name not in dataset and name not in dataset.coords:
                raise ValueError(f'Required variable/coordinate {name} is'
                                 f' not specified in the dataset')

    def __len__(self):
        return int(numpy.prod(self.spectral_values.shape[:-2]))

    def direction_step(self) -> DataArray:
        difference = wrapped_difference(
            numpy.diff(self.direction.values,
                       append=self.direction[0]), period=360
        )
        return DataArray(
            data=difference,
            coords={_NAME_D: self.direction.values},
            dims=[_NAME_D]
        )

    @property
    def radian_direction(self) -> DataArray:
        data_array = self[_NAME_D] * numpy.pi / 180
        data_array.name = 'radian_direction'
        return data_array

    def _directionally_integrate(self, data_array: DataArray) -> DataArray:
        return (data_array * self.direction_step()).sum(_NAME_D)

    @property
    def e(self) -> DataArray:
        return self._directionally_integrate(self[_NAME_E])

    @property
    def a1(self) -> DataArray:
        return self._directionally_integrate(
            self[_NAME_E] * numpy.cos(self.radian_direction)
        )

    @property
    def b1(self) -> DataArray:
        return self._directionally_integrate(
            self[_NAME_E] * numpy.sin(self.radian_direction)
        )

    @property
    def a2(self) -> DataArray:
        return self._directionally_integrate(
            self[_NAME_E] * numpy.cos(2 * self.radian_direction)
        )

    @property
    def b2(self) -> DataArray:
        return self._directionally_integrate(
            self[_NAME_E] * numpy.sin(
                2 * self.radian_direction)
        )

    @property
    def direction(self) -> DataArray:
        return self.dataset[_NAME_D]

    def spectrum_1d(self) -> "FrequencySpectrum":
        return create_1d_spectrum(
            self.frequency.values,
            self.e.values,
            self.time.values,
            self.latitude.values,
            self.longitude.values,
            self.a1.values,
            self.b1.values,
            self.a2.values,
            self.b2.values
        )

    @property
    def number_of_directions(self) -> int:
        return len(self.direction)


class FrequencySpectrum(WaveSpectrum):
    def __init__(self, dataset: Dataset):
        super(FrequencySpectrum, self).__init__(dataset)
        for name in _NAMES_1D:
            if name not in dataset and name not in dataset.coords:
                raise ValueError(f'Required variable/coordinate {name} is'
                                 f' not specified in the dataset')

    def __len__(self):
        return int(numpy.prod(self.spectral_values.shape[:-1]))

    def as_frequency_direction_spectrum(
            self, number_of_directions, method='mem2'
    ) -> "FrequencyDirectionSpectrum":

        direction = numpy.linspace(0, 360, number_of_directions,
                                   endpoint=False)
        radian_direction = direction * numpy.pi / 180

        input_dims = [str(x) for x in self.spectral_values.dims]
        input_shape = self.spectral_values.shape
        output_shape = (*input_shape, number_of_directions)

        loop_elements = numpy.prod(input_shape[:-1])
        output_array = numpy.empty(
            (loop_elements, self.number_of_frequencies, number_of_directions))

        for index in range(loop_elements):
            indices = numpy.unravel_index(index, input_shape[:-1])
            indexers = {
                dim: num for dim, num in zip(input_dims[:-1], indices)
            }
            output_array[index, :, :] = mem2(
                radian_direction,
                self.a1.isel(indexers).values,
                self.b1.isel(indexers).values,
                self.a2.isel(indexers).values,
                self.b2.isel(indexers).values,
            ) * numpy.pi / 180 * self.e.isel(indexers).values[:, None]
        output_array = numpy.reshape(output_array, output_shape)

        return create_2d_spectrum(
            frequency=self.frequency.values,
            direction=direction,
            variance_density=output_array,
            time=self.time.values,
            latitude=self.latitude.values,
            longitude=self.longitude.values,
            depth=self.depth.values
        )


def create_1d_spectrum(
        frequency: numpy.ndarray,
        variance_density: numpy.ndarray,
        time,
        latitude,
        longitude,
        a1=None,
        b1=None,
        a2=None,
        b2=None,
        depth=None) -> FrequencySpectrum:
    time = to_datetime64(numpy.atleast_1d(time))
    latitude = numpy.atleast_1d(latitude)
    longitude = numpy.atleast_1d(longitude)
    if depth is None:
        depth = numpy.zeros(len(time)) + numpy.inf

    if a1 is None: a1 = numpy.nan + numpy.ones_like(variance_density)
    if b1 is None: b1 = numpy.nan + numpy.ones_like(variance_density)
    if a2 is None: a2 = numpy.nan + numpy.ones_like(variance_density)
    if b2 is None: b2 = numpy.nan + numpy.ones_like(variance_density)

    if variance_density.ndim == 1:
        variance_density = variance_density[None, :, ]
        a1 = a1[None, :]
        b1 = b1[None, :]
        a2 = a2[None, :]
        b2 = b2[None, :]

    return FrequencySpectrum(
        Dataset(
            data_vars={
                _NAME_E: ((_NAME_T, _NAME_F), variance_density),
                _NAME_a1: ((_NAME_T, _NAME_F), a1),
                _NAME_b1: ((_NAME_T, _NAME_F), b1),
                _NAME_a2: ((_NAME_T, _NAME_F), a2),
                _NAME_b2: ((_NAME_T, _NAME_F), b2),
                _NAME_LAT: ((_NAME_T,), latitude),
                _NAME_LON: ((_NAME_T,), longitude),
                _NAME_DEPTH: ((_NAME_T,), depth)
            },
            coords={_NAME_T: to_datetime64(time), _NAME_F: frequency}
        )
    )


def create_2d_spectrum(frequency: numpy.ndarray,
                       direction: numpy.ndarray,
                       variance_density: numpy.ndarray,
                       time,
                       latitude,
                       longitude,
                       depth=None) -> FrequencyDirectionSpectrum:
    time = numpy.atleast_1d(time)
    latitude = numpy.atleast_1d(latitude)
    longitude = numpy.atleast_1d(longitude)
    if depth is None:
        depth = numpy.zeros(len(time)) + numpy.inf

    if variance_density.ndim == 2:
        variance_density = variance_density[None, :, :]

    dataset = Dataset(
        data_vars={
            _NAME_E: (
                (_NAME_T, _NAME_F, _NAME_D), variance_density),
            _NAME_LAT: ((_NAME_T,), latitude),
            _NAME_LON: ((_NAME_T,), longitude),
            _NAME_DEPTH: ((_NAME_T,), depth)
        },
        coords={_NAME_T: to_datetime64(time),
                _NAME_F: frequency,
                _NAME_D: direction}
    )
    return FrequencyDirectionSpectrum(dataset)



"""
Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
=========================
Module to define the types used to write read data from Fortran generated.
This is an internal module and there should not be a reason to use this
directly.

Classes:
- `FortranType`, abstract parent class
- `FortranFloat`, class to read Fortran 4 byte floats (float32)
- `FortranInt`, class to read Fortran 4 byte integers (int32)
- `FortranCharacters`, class to read a Fortran character array and return a
    normal python string.

Functions:

- `N/A`

How To Use This Module
=========================
(See the individual functions for details.)

1. Don't :-)
"""
import struct
from io import BufferedIOBase
from typing import Tuple


class FortranType:
    """
    Base class.
    """

    kind = ''  # kind of variable- kind strings are those used by the struct
    # library
    byte_length_type = 1  # bytelength of an individual instance of the data

    def __init__(self, endianness="<"):
        self.endianness = endianness

    def byte_length(self, number: int) -> int:
        """
        bytelength of multiple instances of the data
        :param number: number of type instances
        :return: total lenght in bytes
        """
        return number * self.byte_length_type

    def format(self, number: int) -> str:
        """
        :param number: number of type instanes to read
        :return: format to use with struct unpack to read the fortran data.
        """
        return f"{self.endianness}{self.kind * number}"

    def _read(self, number: int, stream: BufferedIOBase,
              unformatted_sequential=False) -> Tuple[str, bytes]:
        """


        :param number: number of type instanes to read
        :param stream: byte stream
        :param unformatted_sequential: if this is an unformatted sequential
            read
        :return: format to use in unpack and the data (bytes)

        Note on sequential unformatted data:
        https://gcc.gnu.org/onlinedocs/gfortran/
            File-format-of-unformatted-sequential-files.html

        Read sequential unformatted data written by a Fortran program compiled
        with the GFortran compiler. Each record in the file contains a leading
        start of record (sor) 4 byte marker indicating size N of the record,
        the data written (N bytes), and a trailing end of record (eor) marker
        again indicating the size of the data. No information on the type of
        the data is provided, and this has to be known in advance.

        """
        if unformatted_sequential:
            sor = self._read_record_marker(stream)
            number = sor // self.byte_length_type

        data = stream.read(self.byte_length(number))
        fmt = self.format(number)

        if unformatted_sequential:
            eor = self._read_record_marker(stream)
            assert sor == eor
        return fmt, data

    def _unpack(self, number: int, stream: BufferedIOBase,
                unformatted_sequential=False):
        """
        Unpack data
        :param number: number of types to unpack
        :param stream: byte stream
        :param unformatted_sequential: if this is an unformatted sequential
            read
        :return: unpacked data
        """
        return struct.unpack(
            *self._read(number, stream, unformatted_sequential))

    def unpack(self, stream: BufferedIOBase, number=1,
               unformatted_sequential=False):
        """
        Wrapper function. Overwritten by child classes in case some additional
        processing neads to be done before returning the data (e.g. convert
        chart array to str).
        :param stream: byte stream
        :param number: number of types to unpack
        :param unformatted_sequential: if this is an unformatted sequential
            read
        :return: unpacked fortran data as corresponding python type.
        """

        return self._unpack(number, stream, unformatted_sequential)

    def _read_record_marker(self, stream: BufferedIOBase) -> int:
        """
        Read a record marker for an unformatted sequential file.
        :param stream: byte stream
        :return: size of record
        """
        return struct.unpack(
            f"{self.endianness}i", stream.read(4)
        )[0]


class FortranCharacter(FortranType):
    """
    Class to read fortran character types and convert to string.
    """
    kind = 'B'

    def unpack(self, stream: BufferedIOBase, number=1,
               unformatted_sequential=False):
        return "".join([chr(x) for x in
                        self._unpack(number, stream, unformatted_sequential)])


class FortranFloat(FortranType):
    """
    Class to read fortran 32 bit float types and convert to string.
    """
    kind = 'f'
    byte_length_type = 4


class FortranInt(FortranType):
    """
    Class to read fortran 32 bit integer types and convert to string.
    """
    kind = 'i'
    byte_length_type = 4



# Logging
from roguewave.log import logger

# model time
from roguewave.modeldata import (
    TimeSliceLead,
    timebase_forecast,
    TimeSliceForecast,
    TimeSliceEvaluation,
    TimeSliceAnalysis,
    model_timebase,
    model_valid_time,
    available_models,
    list_available_variables,
    open_remote_dataset,
    extract_from_remote_dataset,
    generate_uris
)

# Interpolate
from roguewave.interpolate import (
    interpolate_dataset, tracks_as_dataset,TrackSet,Track,Cluster)

# Wave spectrum
from roguewave.wavespectra import (
    FrequencySpectrum, FrequencyDirectionSpectrum,
    create_1d_spectrum, create_2d_spectrum
)

# IO
from .io.io import save, load

# Restart Files:
from roguewave.wavewatch3 import (
    RestartFile,
    open_restart_file,
    write_partial_restart_file,
    write_restart_file,
    reassemble_restart_file_from_parts
)

# Partitioning
from roguewave.wavespectra.partitioning import (
    get_bulk_partitions_from_spectral_partitions,
    get_spectral_partitions_from_2dspectra,
    get_bulk_partitions_from_observations,
    get_spectral_partitions_from_observations
)

# Colocation
from roguewave.colocate import (
    colocate_model_spotter,
    colocated_tracks_as_dataset,
    colocate_model_spotter_spectra
)






import requests
import os
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

PROD_URL = "https://storm-glass-prod.herokuapp.com/marine-weather/v1"
STAGING_URL = "https://storm-glass-staging.herokuapp.com/marine-weather/v1"
DEV_URL = "https://storm-glass-dev.herokuapp.com/marine-weather/v1"
BASEURL = "https://storm-glass-prod.herokuapp.com/marine-weather/v1"
SOFARURL = "https://api.sofarocean.com/marine-weather/v1"

# configure a session with HTTP retries; https://stackoverflow.com/a/35504626
SESSION = requests.Session()
retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
SESSION.mount("http://", HTTPAdapter(max_retries=retries))
SESSION.mount("https://", HTTPAdapter(max_retries=retries))


def get_token():
    return os.getenv("SOFAR_API_TOKEN", None)


class ExceptionKeyNotFound(Exception):
    pass


def get_url(environment=None):
    if environment:
        if environment == "production":
            if get_token() is None:
                return PROD_URL
            else:
                return SOFARURL
        elif environment == "staging":
            return STAGING_URL
        elif environment == "dev":
            return DEV_URL
    else:
        return BASEURL


def handle_response(response: requests.Response) -> requests.Response:
    if response.status_code == 200:
        return response
    else:
        message = response.content.decode("utf8")
        if "The specified key does not exist" in message:
            raise ExceptionKeyNotFound(f"HTTP response status code {response.status_code} - {message}")
        else:
            raise Exception(f"HTTP response status code {response.status_code} - {message}")


def http_get(endpoint, environment="production"):
    header = {"token": get_token(), "Content-Type": "application/json"}
    tokenstr = f"&token={get_token()}" if get_token() else ""

    url = get_url(environment) + "/" + endpoint + tokenstr
    return handle_response(SESSION.get(url, allow_redirects=True, headers=header))


def set_source(source):
    global BASEURL
    if source == "staging":
        BASEURL = STAGING_URL
    elif source == "production":
        BASEURL = PROD_URL



import numpy
from roguewave.tools.math import wrapped_difference
from roguewave.tools.grid import enclosing_points_1d

def interpolate_periodic(
        xp: numpy.ndarray,
        fp: numpy.ndarray,
        x: numpy.ndarray,
        x_period: float = None,
        fp_period: float = None,
        fp_discont: float = None,
        left: float = numpy.nan,
        right: float = numpy.nan
) -> numpy.ndarray:
    """
    Interpolation function that works with periodic domains and periodic ranges
    :param xp:
    :param fp:
    :param x:
    :param x_period:
    :param fp_period:
    :param fp_discont:
    :return:
    """

    indices =  enclosing_points_1d( xp, x, period=x_period,regular_xp=False )

    delta_x  = wrapped_difference(
        x - xp[indices[0, :]], period=x_period)
    delta_xp = wrapped_difference(
        xp[indices[1,:]] - xp[indices[0,:]],period=x_period)
    delta_fp = wrapped_difference(
        delta=fp[indices[1, :]] - fp[indices[0, :]],
        period=fp_period
    )
    fp0 = fp[indices[0,:]]

    if x_period is not None:
        interpolation = fp0 + delta_fp * delta_x / delta_xp

    else:
        # constant extrapolation
        mask = delta_xp.astype('float64') > 0.0
        interpolation = numpy.zeros( (len(x),) )
        interpolation[mask] =  fp0[mask] + \
                               delta_fp[mask] * delta_x[mask]/delta_xp[mask]
        interpolation[~mask] = fp0[~mask]
        # interpolation = numpy.where( delta_xp.astype('float64') > 0.0, fp0
        #                              + delta_fp * delta_x/delta_xp,fp0)
        # substitute value if given
        if left is not None:
            interpolation = numpy.where( x < xp[0], left, interpolation )

        # substitute value if given
        if right is not None:
            interpolation = numpy.where( x > xp[-1], right, interpolation)

    mask = numpy.isfinite(interpolation)
    interpolation[mask] = wrapped_difference(
        delta=interpolation[mask],
        period=fp_period,
        discont=fp_discont
    )

    return interpolation


def interpolation_weights_1d( xp, x, indices, period: float = None ):
    """
    Find the weights for the linear interpolation problem given a set of
    indices such that:

                xp[indices[0,:]] <= x[:] < xp[indices[1,:]]

    Indices are assumed to be as returned by "enclosing_points_1d" (see
    roguewave.tools.grid).

    Returns weights (nx,2) to perform the one-dimensional piecewise linear
    interpolation to a function given at discrete datapoints xp and evaluated
    at x. Say that at all xp we have for the function values fp,
    the interpolation would then be

        f = fp[ Indices[1,:] ]  *  weights[1,:] +
            fp[ Indices[2,:] ]  *  weights[2,:]

    :param xp:
    :param x:
    :param indices:
    :param period:
    :return:
    """

    weights = numpy.empty_like( indices, dtype='float64' )

    if xp[-1] < xp[0]:
        # make sure we are in a coordinate frame where the vector is
        # in ascending order.
        x =  xp[0] - x
        xp = xp[0] - xp

    delta_x  = wrapped_difference(
        x - xp[indices[0, :]], period=period)

    delta_xp = wrapped_difference(
        xp[indices[1,:]] - xp[indices[0,:]],period=period)

    if period is None:
        mask =  (x >= xp[0]) & (x < xp[-1])
        frac = numpy.empty( (len(x),) )
        frac[mask] = delta_x[mask] / delta_xp[mask]
        frac[ x<xp[0] ] = 1
        frac[x >= xp[-1]] = 0
    else:
        frac = delta_x/delta_xp

    weights[0,:] = 1-frac
    weights[1,:] = frac

    return weights



from datetime import datetime, timezone, timedelta
from pandas import DataFrame
from typing import Union, List, Tuple, TypedDict
import numpy

def get_overlapped_section(model_time: numpy.ndarray,
                           model_variable: numpy.ndarray,
                           observed_time: numpy.ndarray,
                           observed_variable: numpy.ndarray,
                           time_buffer: timedelta = None,
                           wrapping_angle: float = None) -> Union[
    None, Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]]:
    """
    A function that takes two timeseries, a "model" and "observed" series, and
    interpolates them to the same overlapping time interval. Anything outside
    time overlap is discarded unless a buffer is applied- in which case a buffer
    will be added to the left and right of the overlapping interval with missing
    values set to NaN.

    :param model_time: Numpy array of datetime objects
    :param model_variable: 1D numpy array of model values
    :param observed_time: Numpy array of datetime objects
    :param observed_variable: 1D numpy array of observed values
    :param time_buffer: buffer
    :param wrapping_angle: if wrapped variable (e.g. direction) this is the wrapping period.
    :return: If there is overlap: a tuple of three numpy arrays: (
                 time vector (datetime, utc) of overlapping interval,
                 interpolated model values,
                 interpolated time values
                 )
             else:
                 None
    """

    # Pandas does not support negative indices. To ensure this works with
    # pandas series as input use length
    last_element_observed = len(observed_time) - 1
    last_element_modeled = len(model_time) - 1

    # No overlap, return None
    if model_time[0] > observed_time[last_element_observed] or model_time[
        last_element_modeled] < observed_time[0]:
        return None

    # Convert to unix epoch timestamps for interpolation
    model_time = numpy.array([x.timestamp() for x in model_time])
    observed_time = numpy.array([x.timestamp() for x in observed_time])

    # get the start and end time of the overlapping interval
    time_start = model_time[0] if model_time[0] >= observed_time[0] else \
        observed_time[0]
    time_end = model_time[last_element_modeled] if model_time[
                                                       last_element_modeled] <= \
                                                   observed_time[
                                                       last_element_observed] else \
        observed_time[last_element_observed]

    # get the minimum timestep in the output time vector
    delta = min(numpy.min(numpy.diff(model_time)),
                numpy.min(numpy.diff(observed_time)))

    time_end = (int((time_end - time_start) / delta) + 1) * delta + time_start

    # If desired, add a buffer to the left and right of the output interval
    if time_buffer:
        buffer_seconds = delta * (int(time_buffer.total_seconds() / delta))
        time_start -= buffer_seconds
        time_end += buffer_seconds

    # Get the time vector
    time = numpy.linspace(time_start, time_end,
                          int((time_end - time_start) / delta + 1),
                          endpoint=True)

    # If this is a wrapped variable, first unwrap before interpolation
    if wrapping_angle is not None:
        model_variable = numpy.unwrap(model_variable, discont=wrapping_angle)
        observed_variable = numpy.unwrap(observed_variable,
                                         discont=wrapping_angle)

    # Interpolate onto the overlapping time interval
    model_value = numpy.interp(time, model_time,
                               model_variable,
                               left=numpy.nan, right=numpy.nan)
    observed_value = numpy.interp(time, observed_time,
                                  observed_variable,
                                  left=numpy.nan, right=numpy.nan)

    # Convert time back to datetime objects, with UTC timezone
    time_utc = numpy.array(
        [datetime.fromtimestamp(x, tz=timezone.utc) for x in time])

    # If this is a wrapped variable, apply modulo to project back to
    # [0, wrapping_angle] interval
    if wrapping_angle is not None:
        model_value = model_value % wrapping_angle
        observed_value = observed_value % wrapping_angle

    # Return
    return time_utc, model_value, observed_value


def partitions_overlap_in_time(model_time, observed_time) -> bool:
    last_element_observed = len(observed_time) - 1
    last_element_modeled = len(model_time) - 1

    # No overlap, return None
    if model_time[0] > observed_time[last_element_observed] or model_time[
        last_element_modeled] < observed_time[0]:
        return True
    else:
        return False


class MatchOutput(TypedDict):
    model_index: int
    observed_index: int
    model_field: DataFrame
    observed_field: DataFrame
    model_lag: timedelta
    fit: dict


def match(model_fields: List[DataFrame],
          observed_fields: List[DataFrame],
          time_buffer: timedelta = timedelta(hours=24),
          average_time: timedelta = timedelta(hours=24),
          period_threshold=0.8,
          waveheight_threshold=0.8,
          direction_threshold=0.8) -> \
        List[MatchOutput]:
    #
    """
    Try to match wavefields from observed partitions to modeled partitions.
    :param model_fields:
    :param observed_fields:
    :param time_buffer:
    :return:
    """

    # Nested Dict to store matches, first key is the model field index, second key
    # the observed index and the contents are some metrics on the lag and
    # goodness of fit
    matched_to_observation = {}  # [{} for x in observed_fields]

    # Loop over all combinations. To note; fits are _not_ symmetric.
    output = []
    for model_index, model in enumerate(model_fields):
        #
        for observed_index, obs in enumerate(observed_fields):
            #
            # Find the minimum lag between model and observation
            #
            lag = find_model_lag(
                model['time'],
                [model['hm0'], model['tm01'], model['mean_direction']],
                obs['time'],
                [obs['hm0'], obs['tm01'], obs['mean_direction']],
                time_buffer,
                wrapping_angle=[None, None, 360]
            )
            if lag is None:
                # if the lag is None the sections do not overlap
                continue

            variables = ['hm0', 'tm01', 'mean_direction']
            wrapping_angles = [None, None, 360]
            fit = {}
            # Calculate the fit value for waveheight, period and direction once
            # corrected for potential lag
            for variable, wrapping_angle in zip(variables, wrapping_angles):
                time_utc, model_value, observed_value = get_overlapped_section(
                    model['time'] - lag,
                    model[variable],
                    obs['time'],
                    obs[variable],
                    time_buffer=time_buffer,
                    wrapping_angle=wrapping_angle)


                time_delta = time_utc[1] - time_utc[0]
                fit[variable] = time_delta * fit_quality(
                    model_value,observed_value,wrapping_angle) / average_time

            # If the fit value is good enough- we consider it a match
            if fit['hm0'] > waveheight_threshold and fit[
                'tm01'] > period_threshold and fit[
                'mean_direction'] > direction_threshold:

                output.append(MatchOutput(
                    model_index=model_index,
                    observed_index=observed_index,
                    model_field=model_fields[model_index],
                    observed_field=observed_fields[observed_index],
                    model_lag=lag,
                    fit=fit
                ))
    return output


def fit_quality(x, y, wrapping_angle=None):
    if wrapping_angle is not None:
        diff = numpy.abs(
            (x - y + wrapping_angle / 2) % wrapping_angle
            - wrapping_angle / 2
        )
        scale = wrapping_angle / 4
    else:
        diff = numpy.abs(x - y)
        scale = numpy.minimum(x, y)

    return numpy.nansum(1 - numpy.tanh(2 * diff / scale))


def find_model_lag(model_time, model_values, observed_time, observed_values,
                   time_buffer: timedelta,
                   wrapping_angle=None) -> Union[
    None, timedelta]:
    """

    :param model_time:
    :param model_values:
    :param observed_time:
    :param observed_values:
    :param time_buffer:
    :param wrapping_angle:
    :return:
    """
    if not isinstance(model_values, List):
        raise Exception('not a list')

    model = [[] for x in model_values]
    observed = [[] for x in model_values]
    time = None
    if wrapping_angle is None:
        wrapping_angle = [None for x in model_values]

    ii = -1
    for model_variable, observed_variable, wrap in zip(model_values,
                                                       observed_values,
                                                       wrapping_angle):
        out = get_overlapped_section(model_time, model_variable,
                                     observed_time, observed_variable,
                                     time_buffer=time_buffer,
                                     wrapping_angle=wrap)
        ii += 1
        if out:
            time, model[ii], observed[ii] = out
        else:
            return None

    time_delta = (time[2] - time[1])
    nbuffer = int(time_buffer / time_delta)
    out = numpy.zeros(2 * nbuffer + 1)
    for ii in range(-nbuffer, nbuffer + 1):
        for jj in range(len(model)):
            lagged_model = numpy.roll(model[jj], ii)

            if ii > 0:
                lagged_model[:ii] = numpy.nan
            elif ii < 0:
                lagged_model[ii:] = numpy.nan

            out[ii + nbuffer] += fit_quality(observed[jj],
                                             lagged_model,
                                             wrapping_angle[jj])


    index_extreme = numpy.argmax(out)
    time_lag = - time_delta * (index_extreme - nbuffer)

    return time_lag



from roguewave.interpolate.dataset import (
    interpolate_dataset_grid,
    interpolate_dataset_along_axis,
    interpolate_at_points,
    tracks_as_dataset,
    interpolate_dataset
)

from roguewave.interpolate.geometry import (
    TrackSet,Track,Cluster
)



import numpy
from numba import njit, generated_jit, types
GRAV = 9.81

@generated_jit(nopython=True)
def atleast_1d(x):
    if x in types.number_domain:
        return lambda x: numpy.array([x])
    return lambda x: numpy.atleast_1d(x)


@njit(cache=True)
def inverse_intrinsic_dispersion_relation(
    angular_frequency,
    dep,
    grav=GRAV,
    maximum_number_of_iterations=10,
    tolerance=1e-3,
):
    """
    Find wavenumber k for a given radial frequency w using Newton Iteration.
    Exit when either maximum number of iterations is reached, or tolerance
    is achieved. Typically only 1 to 2 iterations are needed.

    :param w: radial frequency
    :param dep: depth in meters
    :param grav:  gravitational acceleration
    :param maximum_number_of_iterations: maximum number of iterations
    :param tolerance: relative accuracy
    :return:
    """

    # Numba does not recognize "atleast_1d" for scalars - hence the weird
    # call to array first.
    w = atleast_1d( angular_frequency)

    k_deep_water_estimate = w ** 2 / grav
    k_shallow_water_estimate = w / numpy.sqrt(grav * dep)

    # == FIRST GUESS==
    # Use the intersection between shallow and deep water estimates to guestimate
    # which relation to use

    k0 = numpy.where( w > numpy.sqrt(grav / dep) ,
                      k_deep_water_estimate,
                      k_shallow_water_estimate)


    # == Newton Iteration ==
    F = numpy.sqrt(k0 * grav * numpy.tanh(k0 * dep)) - w
    cg = numpy.zeros(w.shape, dtype=w.dtype)

    for ii in range(0, maximum_number_of_iterations):
        kd = k0 * dep
        cg = numpy.where(
            kd > 5,
            0.5 * w / k0,
            (1 / 2 + kd / numpy.sinh(2 * kd)) * w / k0
        )
        k0 = k0 - F / cg

        F = numpy.sqrt(k0 * grav * numpy.tanh(k0 * dep)) - w
        error = numpy.abs(F) / w
        if numpy.all(error < tolerance):
            break

    return k0

@njit(cache=True)
def intrinsic_dispersion_relation(k, dep, grav=GRAV):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    k = atleast_1d(k)
    return numpy.sqrt(grav * k * numpy.tanh(k * dep))

@njit(cache=True)
def phase_velocity(k, depth, grav=GRAV):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return intrinsic_dispersion_relation(k, depth, grav=GRAV) / k

@njit(cache=True)
def ratio_group_velocity_to_phase_velocity(k, depth, grav):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    kd = k * depth
    return numpy.where( kd > 5, 0.5, 0.5 + kd / numpy.sinh(2 * kd) )

@njit(cache=True)
def intrinsic_group_velocity(k, depth, grav=GRAV):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return ratio_group_velocity_to_phase_velocity(k, depth, grav=GRAV) * phase_velocity(k, depth, grav)

@njit(cache=True)
def jacobian_wavenumber_to_radial_frequency(k, depth, grav=GRAV):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return 1 / intrinsic_group_velocity(k, depth, grav)

@njit(cache=True)
def jacobian_radial_frequency_to_wavenumber(k, depth, grav=GRAV):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return intrinsic_group_velocity(k, depth, grav)


# Aliasses based on common notation in linear wave theory
c = phase_velocity
cg = intrinsic_group_velocity
k = inverse_intrinsic_dispersion_relation
w = intrinsic_dispersion_relation
n = ratio_group_velocity_to_phase_velocity



from pandas import DataFrame
from roguewave.interpolate.geometry import Geometry
from .timebase import TimeSlice
from roguewave.modeldata.open_remote import open_remote_dataset
from typing import Dict,List, Union
from roguewave.interpolate.dataset import interpolate_dataset
from roguewave.modeldata.remote_point_data import extract_clusters


def extract_from_remote_dataset(
        geometry:Geometry,
        variable:Union[List[str],str],
        time_slice:TimeSlice,
        model_name:str,
        slice_remotely=False,
        parallel=True,
        cache_name=None
) -> Dict[str, DataFrame]:
    """
    Extract timeseries at specified locations from the dataset identified by
    the model name and the timeslice.

    :param geometry: space(time) geometry we want to download from (e.g. a set
        of points, or a spotter track). For most use cases can be specified as

            geometry  =   ( [latitudes ] , [longitudes] )

        more advanced slicing will require the use of interpolation geometries
        defined in roguewave.interpolation.geometry.

    :param variable: name of the variable of interest. Can be a list in which
            case all variables in the list are retrieved.
    :param time_slice: time slice of interest.
    :param model_name: model name
    :param slice_remotely: (default False) if True Skip local cache and try to
    read directly from
        remote. This tries to avoid downloading full files and instead tries
        to only grab the data it needs. Typically slower.
        Does not work for grib files.
    :param parallel: If slice_remotely=True controls whether we download in
        parallel. To note- if parallel is enabled we need to make sure the
        main script is guarded by if __name__ == '__main__'.
    :param cache_name: name of local cache. If None, default cache setup will
        be used.

    :return: Dictionary where each geometry element is a key, and the value is
        a pandas dataframe with as index the time, and as columns the
        requested variables along the given geometry.
    """

    if slice_remotely:
        return extract_clusters(geometry, variable, time_slice,model_name,
                                 parallell=parallel)

    else:
        dataset = open_remote_dataset( variable, time_slice, model_name,
                                       cache_name=cache_name)
        return interpolate_dataset(dataset,geometry)





from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import List
import numpy
from pandas import DataFrame
from roguewave.tools.time import to_datetime_utc

SOFAR_STANDARD_NAMES_WAVE_BULK = {
    "latitude": "latitude",
    "longitude": "longitude",
    "timestamp": "time",
    "significant_waveheight": "significantWaveHeight",
    "peak_period": "peakPeriod",
    "mean_period": "meanPeriod",
    "peak_direction": "peakDirection",
    "peak_directional_spread": "peakDirectionalSpread",
    "mean_direction": "meanDirection",
    "mean_directional_spread": "meanDirectionalSpread",
    "peak_frequency": "peakFrequency"
}


@dataclass
class MetoceanData():
    latitude: float = numpy.nan
    longitude: float = numpy.nan
    time: datetime = datetime(2022, 1, 1, tzinfo=timezone.utc)

    _time: datetime = field(init=False, repr=False)

    @property
    def time(self) -> datetime:
        return self._time

    @time.setter
    def time(self, time):
        if isinstance(time, property):
            # Honestly not sure what is going on- but if we initialze
            # time with the default value of None a "property" object
            # gets passed instead of just "None". This catches that and
            # makes sure it all works. Bit of an edge case, should not
            # happen if initialized with value. This seems to be due to
            # the wave the dataclass and getters and setters interact and
            # the magic involved to make it work.
            time = time.fdel
        self._time = to_datetime_utc(time)

    def _convert(self, data,standard_sofar_names:bool=True ):
        if not standard_sofar_names:
            return data

        out = {}
        for key in data:
            if key in SOFAR_STANDARD_NAMES_WAVE_BULK:
                out[SOFAR_STANDARD_NAMES_WAVE_BULK[key]] = data[key]
            else:
                out[key] = data[key]
        return out

    def as_dict(self,standard_sofar_names:bool=True):
        return self._convert({
            "latitude": self.latitude,
            "longitude": self.longitude,
            "time": self.time
            }, standard_sofar_names)

@dataclass
class WaveBulkData(MetoceanData):

    significant_waveheight: float = numpy.nan
    peak_period: float = numpy.nan
    mean_period: float = numpy.nan
    peak_direction: float = numpy.nan
    peak_directional_spread: float = numpy.nan
    mean_direction: float = numpy.nan
    mean_directional_spread: float = numpy.nan
    peak_frequency: float = numpy.nan

    def as_dict(self,standard_sofar_names:bool=True):
        return self._convert({
            "latitude": self.latitude,
            "longitude": self.longitude,
            "time": self.time,
            "significant_waveheight": self.significant_waveheight,
            "peak_period": self.peak_period,
            "mean_period": self.mean_period,
            "peak_direction": self.peak_direction,
            "peak_directional_spread": self.peak_directional_spread,
            "mean_direction": self.mean_direction,
            "mean_directional_spread": self.mean_directional_spread,
            "peak_frequency": self.peak_frequency
        },standard_sofar_names)


@dataclass
class SSTData(MetoceanData):
    degrees: float = numpy.nan

    def as_dict(self,standard_sofar_names:bool=True):
        return self._convert({
            "latitude": self.latitude,
            "longitude": self.longitude,
            "time": self.time,
            "degrees": self.degrees,
        },standard_sofar_names)

@dataclass
class WindData(MetoceanData):
    speed: float=numpy.nan
    direction: float = numpy.nan
    seasurfaceId: float = numpy.nan

    def as_dict(self,standard_sofar_names:bool=True):
        return self._convert({
            "latitude": self.latitude,
            "longitude": self.longitude,
            "timestamp": self.time,
            "speed": self.speed,
            "direction": self.direction,
            "seasurfaceId": self.seasurfaceId
        },standard_sofar_names)

@dataclass
class BarometricPressure(MetoceanData):
    units: str = ''
    value:float = numpy.nan
    unit_type: str = ''
    data_type_name: str = ''

    def as_dict(self,standard_sofar_names:bool=True):
        return self._convert({
            "latitude": self.latitude,
            "longitude": self.longitude,
            "timestamp": self.time,
            "units": self.units,
            "unit_type": self.unit_type,
            "data_type_name": self.data_type_name,
            "value": self.value
        }, standard_sofar_names)


def as_dataframe(bulk_wave_properties:List[MetoceanData],
                 standard_sofar_names=True):
    data = {}
    for bulk_propererties in bulk_wave_properties:
        dictionary = bulk_propererties.as_dict(standard_sofar_names)
        for key in dictionary:
            if key not in data:
                data[key] = []
            data[key].append( dictionary[key])

    for key in data:
        data[key] = data[key]

    df= DataFrame.from_dict(data)
    if standard_sofar_names:
        df.set_index('time', inplace=True)
    else:
        if 'timestamp' in df:
            df.set_index('timestamp', inplace=True)
        else:
            df.set_index('time', inplace=True)

    return df


"""
Contents: Logic to interact with different type of resources.

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Classes:
- `_RemoteResourceUriNotFound`, exception when a URI does not exist on a
  remote resource.
- `RemoteResource`, abstract base class defining a remote resource
   (s3,http etc)
- `RemoteResourceS3`, class that implements logic to fetch files from s3
- `RemoteResourceHTTPS`, class that implements logic to fetch files using https
"""

from typing import Callable
from boto3 import client
from botocore.exceptions import ClientError
from requests.api import get
from requests.exceptions import HTTPError
from .exceptions import _RemoteResourceUriNotFound
from shutil import copyfile


class RemoteResource:
    """
    Abstract class defining the resource protocol used for remote retrieval. It
    contains just two methods that need to be implemented:
    - download return a function that can download from the resource given a
      uri and filepath
    - method to check if the uri is a valid uri for the given resource.
    """
    URI_PREFIX = 'uri://'

    def download(self) -> Callable[[str, str], bool]:
        """
        Return a function that takes uri (first argument) and filepath (second
        argument), and downloads the given uri to the given filepath. Return
        True on Success. Raise _RemoteResourceUriNotFound if URI does not
        exist on the resource.
        """
        pass

    def valid_uri(self, uri: str) -> bool:
        """
        Check if the uri is valid for the given resource
        :param uri: Uniform Resource Identifier.
        :return: True or False
        """
        if uri.startswith(self.URI_PREFIX):
            return True
        else:
            return False


class RemoteResourceS3(RemoteResource):
    URI_PREFIX = 's3://'

    def __init__(self):
        self.s3 = client('s3')

    def download(self):
        def _download_file_from_aws(uri: str, filepath: str) -> bool:
            """
            Worker function to download files from s3. Raise error if the
            object does not exist on s3.
            :param uri: valid uri for resource
            :param filepath: valid filepath to download remote object to.
            :return: True on success
            """
            s3 = self.s3
            bucket, key = uri.replace(self.URI_PREFIX, '').split('/',
                                                                 maxsplit=1)
            try:
                s3.download_file(bucket, key, filepath)
            except ClientError as e:
                raise _RemoteResourceUriNotFound(
                    f'Error downloading from {uri}. \n'
                    f'Error code: {e.response["Error"]["Code"]} '
                    f'Error message: {e.response["Error"]["Message"]}'
                )
            return True

        return _download_file_from_aws


class RemoteResourceHTTPS(RemoteResource):
    URI_PREFIX = 'https://'

    def download(self):
        def _download_file_from_https(uri: str, filepath: str) -> bool:
            """
            Worker function to download files from https url. Raise error if
            the object does not exist on s3.
            :param uri: valid uri for resource
            :param filepath: valid filepath to download remote object to.
            :return: True on success
            """
            try:
                response = get(uri, allow_redirects=True)
                status_code = response.status_code
                response.raise_for_status()
            except HTTPError as _:
                raise _RemoteResourceUriNotFound(
                    f"Error downloading from: {uri}, "
                    f"http status code: {status_code},"
                    f" message: {response.text}"
                )

            with open(filepath, 'wb') as file:
                file.write(response.content)

            return True

        return _download_file_from_https


class RemoteResourceLocal(RemoteResource):
    URI_PREFIX = 'file://'

    def download(self):
        def _copy_file(uri: str, filepath: str) -> bool:
            """
            Worker function to add local files to a cache. Note that we copy
            to keep the analogy to other remote sources (read only, no changes
            to source).
            :param uri: valid uri for resource
            :param filepath: valid filepath to download remote object to.
            :return: True on success
            """
            source_file = uri.replace(self.URI_PREFIX, '')
            copyfile(source_file, filepath)
            return True

        return _copy_file



"""
Contents: IO

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines that can be used to save and load data.

Functions:

How To Use This Module
======================
(See the individual functions for details.)

"""
from roguewave.wavespectra import FrequencySpectrum, \
    FrequencyDirectionSpectrum, create_1d_spectrum
from roguewave.metoceandata import WaveBulkData, WindData, \
    SSTData,BarometricPressure
from pandas import DataFrame, read_json
from typing import Union, Dict, List
from datetime import datetime
import gzip
import json
import os
import numpy
import base64
from io import BytesIO
from xarray import Dataset, open_dataset


_UNION = Union[
    FrequencySpectrum,
    FrequencyDirectionSpectrum,
    List[FrequencySpectrum],
    Dict[str, List[FrequencyDirectionSpectrum]],
    Dict[str, List[FrequencySpectrum]],
    List[List[FrequencyDirectionSpectrum]],
    List[List[DataFrame]],
    Dict[int,List[DataFrame]],
    Dict[str,List[WaveBulkData]],
    Dict[str,DataFrame]
]

def _b64_encode_numpy(val:numpy.ndarray)->dict:
    if val.dtype == numpy.float64:
        val=val.astype(dtype=numpy.float32)

    if val.dtype == numpy.int64:
        val=val.astype(dtype=numpy.int32)

    dtype = val.dtype.str
    shape = val.shape
    data = base64.b64encode(val)

    return {"__class__":"ndarray", "data":{"dtype":dtype,"shape":shape,"data":data.decode('utf8')}}

def _b64_decode_numpy(data):
    decoded_bytes = base64.b64decode(bytes(data['data'],encoding='utf-8'))
    dtype = numpy.dtype(data['dtype'])
    array = numpy.frombuffer(decoded_bytes,dtype=dtype)
    return numpy.reshape(array,data['shape'])

def _b64_encode_dataset(val:Dataset,name)->dict:
    net_cdf = val.to_netcdf(engine='scipy')
    data = base64.b64encode(net_cdf)
    return {"__class__":name, "data":{"data":data.decode('utf8')}}

def _b64_decode_dataset(data):
    decoded_bytes = base64.b64decode(bytes(data['data'],encoding='utf-8'))
    with BytesIO(decoded_bytes) as fp:
        dataset = open_dataset(fp,engine='scipy')
    return dataset

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, numpy.ndarray):
            return _b64_encode_numpy(obj)
        elif isinstance(obj, numpy.int32):
            return int(obj)
        elif isinstance(obj, numpy.float32):
            return float(obj)
        elif isinstance(obj, numpy.int64):
            return int(obj)
        elif isinstance(obj, numpy.float64):
            return float(obj)
        elif isinstance(obj,Dataset):
            return _b64_encode_dataset(obj,"Dataset")
        elif isinstance(obj,FrequencySpectrum):
            return _b64_encode_dataset(obj.dataset,"FrequencySpectrum")
        elif isinstance(obj,FrequencyDirectionSpectrum):
            return _b64_encode_dataset(obj.dataset,"FrequencyDirectionSpectrum")
        elif isinstance(obj, datetime):
            return {'__class__':'datetime', 'data':datetime.isoformat(obj)}
        elif isinstance(obj,DataFrame):
            return {'__class__':'DataFrame', 'data': obj.to_json(date_unit='s')}
        elif isinstance(obj, WaveBulkData):
            return {'__class__': 'WaveBulkData', 'data':obj.as_dict()}
        elif isinstance(obj, WindData):
            return {'__class__': 'WindData', 'data':obj.as_dict()}
        elif isinstance(obj, SSTData):
            return {'__class__': 'SSTData', 'data':obj.as_dict()}
        elif isinstance(obj, BarometricPressure):
            return {'__class__': 'BarometricPressure', 'data':obj.as_dict()}
        else:
            return json.JSONEncoder.default(self, obj)

def object_hook(dictionary:dict):
    if '__class__' in dictionary:
        if dictionary['__class__'] == 'DataFrame':
            #
            df = read_json(dictionary['data'])
            if 'timestamp' in df:
                df['timestamp'] = df['timestamp'].apply(
                    lambda x: x.tz_localize('utc'))
            elif 'time' in df:
                df['time'] = df['time'].apply(
                    lambda x: x.tz_localize('utc'))
            elif 'valid_time' in df:
                df['valid_time'] = df['valid_time'].apply(
                    lambda x: x.tz_localize('utc'))
            elif 'init_time' in df:
                df['valid_time'] = df['valid_time'].apply(
                    lambda x: x.tz_localize('utc'))
            else:
                df.index = [ x.tz_localize('utc') for x in df.index]
            return df
        elif dictionary['__class__'] == 'WaveSpectrum1D':
            data = dictionary['data']
            return create_1d_spectrum(
                frequency=data['frequency'],
                variance_density=data['varianceDensity'],
                time=data['timestamp'],
                latitude=data['latitude'],
                longitude=data['longitude'],
                a1=data['a1'],
                b1=data['b1'],
                a2=data['a2'],
                b2=data['b2']
            )
        elif dictionary['__class__'] == 'Dataset':
            return _b64_decode_dataset(dictionary['data'])
        elif dictionary['__class__'] == 'FrequencySpectrum':
            return FrequencySpectrum(
                _b64_decode_dataset(dictionary['data']))
        elif dictionary['__class__'] == 'FrequencyDirectionSpectrum':
            return FrequencyDirectionSpectrum(
                _b64_decode_dataset(dictionary['data']))
        elif dictionary['__class__'] == 'ndarray':
            return _b64_decode_numpy(dictionary['data'])
        elif dictionary['__class__'] == 'datetime':
            return datetime.fromisoformat( dictionary['data'] )
        elif dictionary['__class__'] == 'WaveBulkData':
            return WaveBulkData(**dictionary['data'])
        elif dictionary['__class__'] == 'WaveBulkData':
            return WaveBulkData(**dictionary['data'])
        elif dictionary['__class__'] == 'WindData':
            return WindData(**dictionary['data'])
        elif dictionary['__class__'] == 'SSTData':
            return SSTData(**dictionary['data'])
        elif dictionary['__class__'] == 'BarometricPressure':
            return BarometricPressure(**dictionary['data'])
    else:
        return dictionary

def load(filename: str) -> _UNION:
    """
    Load spectral data as saved by "save_spectrum" from the given file and
    return a (nested) object. The precise format of the output depends on what
    was saved.

    :param filename: path to file to load.
    :return:
        - Data in the same form it was saved.
    """

    with gzip.open(filename, 'rb') as file:
        data = file.read().decode('utf-8')
        return json.loads(data, object_hook=object_hook)


def save(_input: _UNION, filename: str, format='json', separate_spotters_into_files=False):
    """
    Save spectral data, possible in nested form as returned by the spectral
    partition/reconstruction/etc. functions. Data is saved in JSON form
    compressed with gzip.

    :param _input:
        - Data containing python primitives (dict/list) or any of the RogueWave
          classes.

    :param filename: path to save data.
    :return: None
    """

    def write(filename, _input, format):
        if format == 'json':
            with gzip.open(filename, 'wt') as file:
                json.dump(_input,file,cls=NumpyEncoder)
                #file.write(data.encode('utf-8'))
        else:
            raise Exception('Unknown output format')


    if separate_spotters_into_files:
        os.makedirs(filename, exist_ok=True)
        for key in _input:
            name = os.path.join(filename, key)
            write( name,_input[key],format )

    else:
        write(filename,_input,format)


from roguewave.wavewatch3 import RestartFileStack, open_restart_file_stack
from .timebase import TimeSlice
from .keygeneration import generate_uris


def open_remote_restart_file(
        variable: str,
        time_slice: TimeSlice,
        model_name: str,
        model_definition_file: str,
        cache=True,
        cache_name: str = None )-> RestartFileStack:

    aws_keys = generate_uris(variable, time_slice, model_name)
    return open_restart_file_stack(aws_keys,
                                   model_definition_file,cache,cache_name )


from pysofar.sofar import SofarApi
from typing import List

# SofarAPI instance.
_API = None


# Helper functions
def get_spotter_ids(sofar_api: SofarApi = None) -> List[str]:
    """
    Get a list of Spotter ID's that are available through this account.

    :param sofar_api: valid SofarApi instance.
    :return: List of spotters available through this account.
    """
    if sofar_api is None:
        sofar_api = _get_sofar_api()
    return sofar_api.device_ids
# -----------------------------------------------------------------------------


def _get_sofar_api() -> SofarApi:
    """
    Gets a new sofar API object if requested. Returned object is essentially a
    Singleton class-> next calls will return the stored object instead of
    creating a new class. For module internal use only.

    :return: instantiated SofarApi object
    """
    global _API
    if _API is None:
        _API = SofarApi()
    return _API
# -----------------------------------------------------------------------------


from roguewave.wavespectra import FrequencySpectrum, create_1d_spectrum
from datetime import datetime
from pandas import read_csv, to_numeric
import numpy
import os

def get_spectrum_from_parser_output(path: str)->FrequencySpectrum:
    """

    :param path: Path that contains the output from the spotter parser.
    :return: A list of WaveSpectrum1D objects.
    """
    def load_spectral_file(file):
        if os.path.isfile(file):
            data = read_csv(file).apply(to_numeric, errors='coerce')
        else:
            raise FileNotFoundError(file)

        columns = list(data.columns)
        frequencies = [float(x) for x in columns[8:]]
        values = data[columns[8:]].values

        time_tuple = data[columns[0:6]].values
        time = []
        for index in range(time_tuple.shape[0]):
            time.append(
                datetime(year=time_tuple[index, 0], month=time_tuple[index, 1],
                         day=time_tuple[index, 2], hour=time_tuple[index, 3],
                         minute=time_tuple[index, 4],
                         second=time_tuple[index, 5]))

        return {'time': numpy.array(time), 'frequencies': frequencies,
                'values': values}

    files = ['a1', 'b1', 'a2', 'b2', 'Szz']
    data = {}
    for file_type in files:
        file_location = os.path.join(path, file_type+'.csv')
        data[file_type] = load_spectral_file(file_location)

    number_of_spectra = data['Szz']['values'].shape[0]
    spectra = []
    for index in range(0,number_of_spectra):
        spectra.append(
            create_1d_spectrum(
                frequency=data['Szz']['frequencies'],
                variance_density=data['Szz']['values'][index,:],
                time=data['Szz']['time'][index],
                latitude=None,
                longitude=None,
                a1=data['a1']['values'][index,:],
                b1=data['b1']['values'][index,:],
                a2=data['a2']['values'][index,:],
                b2=data['b2']['values'][index,:]
            )
        )
    return FrequencySpectrum.concat_from_list(spectra)


"""
This file is part of pysofar: A client for interfacing with Sofar Oceans Spotter API

Contents: Wind Estimater

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""
import typing
import numpy

def U10( E , f , a1=None , b1=None  ,fmin=-1., fmax=.5  ,Npower=4        , I=2.5 ,
               beta=0.012, kappa=0.4,alpha=0.012, grav=9.81, numberOfBins=20)->typing.Tuple[float,float,float]:
    #
    # =========================================================================
    # Required Input
    # =========================================================================    
    # 
    # f              :: frequencies (in Hz)
    # E              :: Variance densities (in m^2 / Hz )
    #
    # =========================================================================
    # Output
    # =========================================================================    
    #
    # U10            :: in m/s
    # Direction      :: in degrees clockwise from North (where wind is *coming from)
    #
    # =========================================================================
    # Named Keywords (parameters to inversion algorithm)
    # =========================================================================
    # Npower = 4     :: exponent for the fitted f^-N spectral tail
    # I      = 2.5   :: Philips Directional Constant
    # beta   = 0.012 :: Equilibrium Constant
    # Kapppa = 0.4   :: Von Karman constant
    # Alpha  = 0.012 :: Constant in estimating z0 from u* in Charnock relation
    # grav   = 9.81  :: Gravitational acceleration
    #
    # =========================================================================
    # Algorithm
    # =========================================================================    
    #
    # 1) Find the part of the spectrum that best fits a f^-4 shape
    # 2) Estimate the Phillips equilibrium level "Emean" over that range
    # 3) Use Emean to estimate Wind speed (using Charnock and LogLaw)   
    # 4) Calculate mean direction over equilibrium range
    
    
    
    # If fmin smaller than 0; use fraction of peak frequency as minimum
    if fmin < 0.:
        #
        iMin = numpy.argmax( E )
        fmin = numpy.abs(fmin) * f[iMin]
        #
    #
    
    #
    # Find the equilibrium range; output is a list of indices of the spectrum array that make up the tail
    #
    equilibriumRange = findEquilibriumRange( E , f ,  fmin=fmin , fmax=fmax ,  Npower=Npower, numberOfBins = numberOfBins)    
    #
    # Caclulate the weighted mean over the tail
    
    Emean            = 8. * numpy.pi**3 * numpy.mean( E[equilibriumRange] * f[equilibriumRange]**Npower )
        
    # Get friction velocity from spectrum
    Ustar    = Emean / grav / I / beta / 4

    # Find z0 from Charnock Relation
    z0       = alpha * Ustar**2 / grav
        
    # Get the wind speed at U10 from loglaw
    U  = Ustar / kappa * numpy.log( 10. / z0 )        
    
    #
    if a1 is not None:
        #
        # Calculate the mean directional moments over the equilibrium range.
        #
        a1m = numpy.mean( a1[equilibriumRange] )
        b1m = numpy.mean( b1[equilibriumRange] )   
        #
        # Convert directions to where the wind is coming from, measured positive clockwise from North
        dir = (270. - 180. / numpy.pi * numpy.arctan2( b1m , a1m )) % 360
                #
    else:
        #
        dir = None
        #
    #
    return( U , dir, Ustar )
    #
#enddef

def findEquilibriumRange( E , f ,  fmin=0.0293 , fmax=1.25 ,  Npower=4, numberOfBins = 20):
    #
    # fmin        :: minimum frequency considered (default is Spotter minimum)
    # fmax        :: maximum frequency considered (default is Spotter Maximum)
    import numpy

    #Get power spectrum
    E = numpy.squeeze(  E * f ** Npower )
    
    #Find fmin/fmax
    iMin = numpy.argmin( numpy.abs( f - fmin ) ,axis= -1 )
    iMax = numpy.argmin( numpy.abs( f - fmax ) ,axis= -1 )
    nf   = len(f)
    
    numberOfFrequencies   = E.shape[-1]
    iMax                  = iMax + 1 - numberOfBins
    iMax                  = numpy.max( (iMin + 1, iMax) )
    iMax                  = numpy.min( (iMax , nf-numberOfBins) )
    
    iCounter = 0
    Variance = numpy.zeros( iMax-iMin ) + 1.e10
    #
    # Calculate the variance with respect to a running average mean of numberOfBins
    #
    for iFreq in range( iMin , iMax ):
        #        
        #Ensure we do not go out of bounds
        iiu = numpy.min( (iFreq+numberOfBins, nf )  )
        
        #Ensure there are no 0 contributions (essentially no data)
        msk = E[iFreq:iiu] > 0.
        if (numpy.sum(msk) == numberOfBins):
            M                  = numpy.mean(  E[iFreq:iiu]         )
            Variance[iCounter] = numpy.mean( (E[iFreq:iiu] - M)**2 ) / ( M**2 )
            
        iCounter = iCounter + 1        
        #
    #
    iMinVariance = numpy.argmin( Variance )
    iiu = numpy.min( (iMinVariance + iMin  + numberOfBins, nf ) )
    #
    return ( numpy.arange( iMinVariance + iMin, iiu ) )
    #
#enddef


"""
Contents: Simple local caching of spotter requests. We simply add a wavefleet
remote resource to a file cache and add custom routines to create and retrieve
data from a cache.

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Classes:
- `WaveFleetResource`, remote resource specification for grabbing data from
   wavefleet

Functions:



Use:

"""

# Import
# =============================================================================


from roguewave.filecache.remote_resources import RemoteResource
from roguewave import filecache
from roguewave import save, load
from roguewave.io.io import NumpyEncoder, object_hook
import json
from typing import List, Callable
from pysofar.spotter import SofarApi


# Constants
# =============================================================================


# Spotter Cache setup.
CACHE_NAME = 'spotter_cache'
CACHE_PATH = '~/temporary_roguewave_files/spotter_cache'
CACHE_SIZE_GB = 2


# Classes
# =============================================================================


class WaveFleetResource(RemoteResource):
    """
    Remote resource for downloading data from wavefleet given a "URI".
    The URI in this case is merely a JSON encoded version of the kwargs to
    a call to the appropriate Spotter API download functions. The request type
    and function pointer are registerd as keyword/value pairs in the
    _handlers dictionary.
    """
    URI_PREFIX = 'wavefleet://'

    def __init__(self,
                 request_type_handle_mapping: dict,
                 session: SofarApi):
        # Dictionary that maps a request type (str) to a function
        self._handlers = request_type_handle_mapping
        # Sofar session object
        self.session = session

    def download(self) -> Callable[[str, str], bool]:
        # download function conforming to RemoteResource protocol of FileCache.
        def get_data_from_wavefleet(uri: str, filepath: str) -> bool:
            request_type, kwargs = _decode_spotter_cache_uri(uri)
            data = self._handlers[request_type](session=self.session, **kwargs)
            save(data, filepath)
            return True

        return get_data_from_wavefleet

    def handler_defined(self, request_type) -> bool:
        return request_type in self._handlers

    def add_handler(self, request_type, handler):
        if not self.handler_defined(request_type=request_type):
            self._handlers[request_type] = handler
        else:
            raise ValueError(
                f'A handler is already defined for {request_type}')


# Functions
# =============================================================================


def get_data(spotter_ids: List[str], session: SofarApi,
             handler, **kwargs):
    """
    Caching method for the get_data function defined in spotter_api. Note that
    the function handles creation of a cache if not initialized and adding of
    the request handler.

    :param spotter_ids: list of spotter ID's
    :param session: valid sofar API session
    :param handler: Function to execute a given request.
    :param kwargs: Args needed to call the handler.
    :return:
    """
    REQUEST_TYPE = 'get_data'
    if not _exists():
        _create_cache(request_type=REQUEST_TYPE, handler=handler,
                      session=session)

    if not _handler_is_defined(request_type=REQUEST_TYPE):
        _add_handler(request_type=REQUEST_TYPE, handler=handler)

    uris = [_spotter_cache_uri(REQUEST_TYPE, spotter_id=_id,
                               **kwargs) for _id in spotter_ids]

    filepaths = filecache.filepaths(uris, cache_name=CACHE_NAME)
    output = [load(filepath) for filepath in filepaths]
    return output


def get_data_search(handler, session: SofarApi, **kwargs):
    """
    Caching method for the search functions defined in spotter_api. Note that
    the function handles creation of a cache if not initialized and adding of
    the request handler.

    :param handler: Function to execute a given request.
    :param session: valid sofar API session
    :param kwargs: Args needed to call the handler.

    :return:
    """
    REQUEST_TYPE = 'search'
    if not _exists():
        _create_cache(request_type=REQUEST_TYPE, handler=handler,
                      session=session)

    if not _handler_is_defined(request_type=REQUEST_TYPE):
        _add_handler(request_type=REQUEST_TYPE, handler=handler)

    uri = _spotter_cache_uri(REQUEST_TYPE, **kwargs)
    filepath = filecache.filepaths(uri, cache_name=CACHE_NAME)[0]
    return load(filepath)


# Module internal functions
# =============================================================================


def _spotter_cache_uri(request_type: str, **kwargs):
    """
    Encode a set of keyword arguments as a wavefleet "URI". The wavefleet uri
    is a string that has the form:

        wavefleet://[request_type]/[json_serialized_kwargs]

    To note; this is only used to create a hashable string that uniquely
    specifies the request.

    :param request_type: request type unique identifier that has a
        corresponding handling function in the WaveFleetResource
    :param kwargs: json serializabe dict. See NumpyEncoder for custom additions
    :return:  uri
    """
    kwargs['request_type'] = request_type
    return f"wavefleet://" + json.dumps(kwargs, cls=NumpyEncoder)


def _decode_spotter_cache_uri(uri: str):
    """
        Decode a "wavefleet uri" back into its kwargs form together with
        the request type. A "wavefleet uri" is a string that has the form:

            wavefleet://[request_type]/[json_serialized_kwargs]

        To note; this is only used to create a hashable string that uniquely
        specifies the request.

        :param uri: wavefleet "URI"
        :return:  request_type and kwargs that serve as input to the
            request_type handler.
        """
    uri = uri.replace('wavefleet://', '')
    kwargs = json.loads(uri, object_hook=object_hook)
    request_type = kwargs.pop('request_type')
    return request_type, kwargs


def _create_cache(request_type, handler, session):
    if not _exists():
        wavefleetresource = WaveFleetResource({request_type: handler}, session)
        filecache.create_cache(CACHE_NAME,
                               cache_path=CACHE_PATH,
                               resources=[wavefleetresource],
                               cache_size_GB=CACHE_SIZE_GB)


def _exists() -> bool:
    """
    Check if a spotter cache already exists
    :return: True or False
    """
    return filecache.exists(CACHE_NAME)


def _add_handler(request_type, handler):
    """
    Add a handler to the WaveFleetResource object of the spotter file cache
    that handles the given request type. If a handler with the given
    request type already exists we do _NOT_ update.

    :param request_type: request type unique identifier
    :param handler: function that handles kwargs associated with the request
        type
    :return:
    """

    cache = filecache.get_cache(CACHE_NAME)
    # Remote resources are stored in the resource list of the cache object.
    for remote_resource in cache.resources:
        # We need to get the WaveFleetResource Object to add the handler too
        if isinstance(remote_resource, WaveFleetResource):
            remote_resource.add_handler(request_type, handler)
            break
    else:
        # Note that we require that the wavefleet object is there. If the
        # cache was created through create_cache this should be the case.
        raise ValueError('No valid WaveFleetResource object instantiated')


def _handler_is_defined(request_type):
    """
    Check if a handler is defined for the request type.
    :param request_type: request type unique identifier.
    :return:
    """
    cache = filecache.get_cache(CACHE_NAME)
    for remote_resource in cache.resources:
        if isinstance(remote_resource, WaveFleetResource):
            return remote_resource.handler_defined(request_type)
    else:
        # Note that we require that the wavefleet object is there. If the
        # cache was created through create_cache this should be the case.
        raise ValueError('No valid WaveFleetResource object instantiated')



import boto3

def aws_byte_range(bucket, key,byterange):
    obj = boto3.resource('s3').Object(bucket, key)
    stream = obj.get(Range=f'bytes={byterange[0]}-{byterange[1]}')['Body']
    return stream.read()



"""
Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Module defining the various resources we can use. Here a resource is a class
that acts as a normal file opened for binary read or write in Python. However,
the underlying object may be a an s3 object or a local file. In the case of
an s3 object all IO is buffered and changes occur on s3 only once the file
object is closed.

Classes:
- `Resource`, abstract parent class
- `FileResource`, resource implementation when we use local files (basically
    wrapper around a standard file object)
- `S3Resource`, resource implementation when we use objects stored on S3.

Functions:

- `create_resource`, create a resource, main function to use

How To Use This Module
======================
(See the individual functions for details.)

1. `import create_resource from resources`
2. create a resource: resource = create_resource( uri,mode ), where uri is
    either a local file path, or a s3 uri, and mode is read binary ('rb') or
    write binary ('wb').
"""
from multiprocessing.pool import ThreadPool
from multiprocessing import Lock
from typing import Union, Sequence, List, Literal
from io import BytesIO
from boto3 import resource
from tqdm import tqdm
from roguewave.wavewatch3.restart_file_cache import get_data


class Resource():
    """
    Context manager base clase defining a resource. needs to be implemented.
    """
    def __init__(self,
                 resource_location,
                 mode:Literal['rb','wb']='rb',
                 cache:bool=False,
                 cache_name: str = None
                 ):
        """
        :param resource_location:
        :param mode: mode to open file, only binary read ('rb') or binary write
            ('wb') are supported.
        """
        self.resource_location = resource_location
        self.mode = mode
        self.cache = cache
        self.cache_name = cache_name

    @property
    def read_only(self):
        """
        :return: resource is opened as readonly
        """
        return self.mode == 'rb'

    @property
    def write_only(self):
        """
        :return: resource is opened as writeonly
        """
        return self.mode == 'wb'

    def read_range(self, s:Union[slice,Sequence[slice]]
                   ) -> List[bytearray]:
        """
        Read a range of bytes as defined by the slice(s). We can
        input a single slice, or multiple slices. We always return a list of
        bytearrays corresponding to the byteranges (list of length 1 if a
        single slice is used as input

        :param s: slice or list of slices
        :return: List of bytearrays
        """

        pass

    def read(self, number_of_bytes=-1) -> bytearray:
        """
        :param number_of_bytes: read number of bytes, if <0 read all bytes to
            end of resource from current posiiton
        :return: return a bytearray of the data read
        """
        pass

    def write(self, _bytes:bytes):
        """
        write a byte array to the resource
        :param _bytes: bytes to write
        :return: None
        """
        pass

    def __enter__(self):
        """
        Dunder method to implement the context management protocol on entry of
        a with block.
        :return:
        """
        return self

    def close(self):
        """
        Close the resource
        :return: None
        """
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Dunder method to implement the context management protocol on entry of
        a with block.
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self.close()
        self.resource_handle = None

    def __del__(self):
        """
        Dunder method to make sure we release any resources properly.
        :return:
        """
        if self.resource_handle is not None:
            self.close()

    def seek(self, position_bytes):
        """
        Move the current position in the stream to the indicated position.
        :param position_bytes: position in bytes
        :return:
        """
        pass

    def tell(self):
        """
        Return the current position in the stream.
        :return: position in stream in bytes.
        """
        pass


class FileResource(Resource):
    """
    Open a file as a "resource". Only needed becuase we want to abstract away
    the difference between a local and remote file for s3.
    """
    def __init__(self,
                 resource_location,
                 mode:Literal['rb','wb']='rb',
                 cache:bool=False,
                 cache_name: str = None):
        """
        :param resource_location:
        :param mode: mode to open file, only binary read ('rb') or binary write
            ('wb') are supported.
        """
        super().__init__(resource_location, mode, False, None)
        self.resource_handle = open(resource_location, mode)

    def read_range(self, slices:Union[slice,Sequence[slice]]) \
            -> Union[bytearray,List[bytearray]]:
        """
        Read a range of bytes as defined by the slice(s). We can
        input a single slice, or multiple slices. We always return a list of
        bytearrays corresponding to the byteranges (list of length 1 if a
        single slice is used as input

        :param s: slice or list of slices
        :return: List of bytearrays
        """


        if self.write_only:
            raise IOError('Resource has been opened as write only, cannot'
                             'read from it.')

        if isinstance(slices,slice):
            slices = [slices]

        out = []
        for s in slices:
            self.resource_handle.seek(s.start)
            number_bytes = s.stop - s.start
            out.append(bytearray(self.resource_handle.read(number_bytes)))

        return out

    def write(self,_bytes:bytes):
        """
        write a byte array to the resource
        :param _bytes: bytes to write
        :return: None
        """
        self.resource_handle.write(_bytes)

    def read(self, number_of_bytes=-1) -> bytearray:
        """
        :param number_of_bytes: read number of bytes, if <0 read all bytes to
            end of resource from current posiiton
        :return: return a bytearray of the data read
        """
        if self.write_only:
            raise IOError('Resource has been opened as write only, cannot'
                             'read from it.')

        return bytearray(self.resource_handle.read(number_of_bytes))

    def close(self):
        """
        Close the resource
        :return: None
        """
        if self.resource_handle is not None:
            self.resource_handle.close()

    def seek(self, position_bytes):
        """
        Move the current position in the stream to the indicated position.
        :param position_bytes: position in bytes
        :return:
        """
        return self.resource_handle.seek(position_bytes)

    def tell(self):
        """
        Return the current position in the stream.
        :return: position in stream in bytes.
        """
        return self.resource_handle.tell()


class S3Resource(Resource):
    """
    Open a s3 object as a "resource". Allows us to interact with an s3 object
    as if it was a regular stream (supports read, write, seek, tell). Obviously
    less performant than local IO.
    """
    boto3_client_lock = Lock()

    def __init__(self, resource_location, mode:Literal['rb','wb']='rb',
                 cache:bool=False, cache_name: str = None):
        """
        :param resource_location:
        :param mode: mode to open file, only binary read ('rb') or binary write
            ('wb') are supported.
        """
        super().__init__(resource_location,mode,cache,cache_name)

        bucket, key = \
            resource_location.replace('s3://', '').split('/', maxsplit=1)
        self._key = key
        self._bucket = bucket

        # Creation is not thread safe. Ensure only one thread is creating.
        with self.boto3_client_lock:
            self.resource_handle = resource('s3')
        self._position = 0
        self._write_buffer = BytesIO()
        self._bytes_written = False


    def read_range(self, slices: Union[slice, Sequence[slice]]) \
            ->List[bytearray]:
        """
        Read a range of bytes as defined by the slice(s). We can
        input a single slice, or multiple slices. We always return a list of
        bytearrays corresponding to the byteranges (list of length 1 if a
        single slice is used as input.

        Note that for efficiency we use multiple threads to iterate over the
        byterange. Only the byterange needed is downloaded.

        :param s: slice or list of slices
        :return: List of bytearrays
        """
        if isinstance(slices,slice):
            slices = [slices]

        if self.cache:
            keys = []
            start_byte = []
            stop_byte = []
            for _slice in slices:
                stop_byte.append(_slice.stop)
                start_byte.append(_slice.start)
                keys.append(self.resource_location)
            return get_data(keys,start_byte,stop_byte,self.cache_name)

        if self.write_only:
            raise IOError('Resource has been opened as write only, cannot'
                             'read from it.')

        def _worker( s:slice):
            if s.start == 0 and s.stop == -1:
                obj = self.resource_handle.Object(self._bucket, self._key)
                with BytesIO() as file:
                    response = obj.download_fileobj(file)
                    file.seek(0)
                    _bytes = bytearray(file.read())

                return bytearray(_bytes)
            else:
                obj = self.resource_handle.Object(self._bucket, self._key)
                data = obj.get(Range=f'bytes={s.start}-{s.stop-1}')['Body']
                return bytearray(data.read())



        if len(slices) > 1:
            with ThreadPool(processes=10) as pool:
                output = list(
                    tqdm(
                        pool.imap(_worker, slices),
                        total=len(slices)
                    )
                )
        else:
            output = [_worker(slices[0])]
        return output


    def read(self, number_of_bytes=-1) -> bytearray:
        """
        :param number_of_bytes: read number of bytes, if <0 read all bytes to
            end of resource from current posiiton
        :return: return a bytearray of the data read
        """
        if self.write_only:
            raise IOError('Resource has been opened as write only, cannot'
                             'read from it.')

        start_byte = self._position
        if number_of_bytes < 0:
            end_byte = -1
            _bytes= self.read_range(slice(0,end_byte,1 ))[0]
            return _bytes[start_byte:]
        else:
            end_byte = self._position + number_of_bytes
            self._position += number_of_bytes
            return self.read_range(slice( start_byte,end_byte,1 ))[0]

    def write(self, _bytes:bytes):
        """
        Write a byte array to the resource. Note that the s3 resource writes
        to an internal buffer. Only when the object is closed is the buffer
        pushed to s3 and "written".
        :param _bytes: bytes to write
        :return: None
        """

        self._write_buffer.write(_bytes)
        self._bytes_written = True

    def _push_to_s3(self):
        """
        Push data in the write buffer to s3.
        :return:
        """

        self._write_buffer.seek(0)
        obj = self.resource_handle.Object(self._bucket, self._key)
        response = obj.upload_fileobj(self._write_buffer)

    def close(self):
        """
        Close the resource. Flushes the write buffer (if applicable).
        :return: None
        """
        if self._bytes_written:
            self._push_to_s3()
        self._write_buffer.close()
        self._bytes_written = False

    def tell(self):
        """
        Return the current position in the stream.
        :return: position in stream in bytes.
        """
        return self._position

    def seek(self, position_bytes):
        """
        Move the current position in the stream to the indicated position.
        :param position_bytes: position in bytes
        :return:
        """
        self._position = position_bytes


def create_resource(uri:str,mode:Literal['wb','rb']='rb',cache=False,
                    cache_name=None)->Resource:
    """
    Create the appropriate resource object basded on the "uri". If the uri
    starts with "s3://" we assume it refer to a s3 object in the form
    "s3://bucket/key". Otherwise we assume it is the path to a local file.
    :param uri: s3 uri or path to local file.
    :param mode: Whether to open resource in binary read ('rb') or binary write
        mode.
    :return: resource object
    """
    if 's3://' in uri:
        return S3Resource(resource_location=uri,mode=mode,
                          cache=cache,cache_name=cache_name)
    else:
        return FileResource(resource_location=uri,mode=mode,
                            cache=cache,cache_name=cache_name)


import numpy
from dataclasses import dataclass
from typing import Sequence, Mapping, List, Union
from pandas import DataFrame
from roguewave.tools.time import to_datetime64, to_datetime_utc
from datetime import datetime
from roguewave.interpolate.general import interpolate_periodic
from numbers import Number
from xarray import Dataset

class _Geometry:
    pass

@dataclass()
class SpatialPoint(_Geometry):
    latitude: float
    longitude: float
    id:str

    @property
    def is_valid(self)->bool:
        return numpy.isfinite(self.latitude) and numpy.isfinite(self.longitude)

@dataclass()
class SpaceTimePoint(SpatialPoint):
    time: datetime

    @classmethod
    def from_spatial_point(cls,point:SpatialPoint,time:datetime):
        return cls( point.latitude,point.longitude,point.id,time )


@dataclass()
class Cluster(_Geometry):
    """
    A cluster is a set of points, each identified by unique id.
    """
    points: Mapping[str,SpatialPoint]

    @classmethod
    def from_lat_lon_arrays(cls, lats, lons):
        points = {}
        for index in range(0, len(lats)):
            _id = str(index)
            points[_id] = SpatialPoint(lats[index], lons[index],_id)

        return cls(points)

    @property
    def latitude(self) -> numpy.ndarray:
        return numpy.array([x.latitude for x in self.points.values()])

    @property
    def longitude(self) -> numpy.ndarray:
        return numpy.array([x.longitude for x in self.points.values()])

    @property
    def ids(self) -> Sequence[str]:
        return [x for x in self.points.keys()]

@dataclass()
class ClusterStack(_Geometry):
    """
    A cluster timestack is a stack of clusters in time, e.g. a cluster of
    spotters as it evolves in time.
    """
    time:numpy.ndarray
    clusters:Sequence[Cluster]

    def as_track_set(self) -> "TrackSet":
        return TrackSet.from_clusters(self)

    @classmethod
    def from_track_set(cls, track_set:"TrackSet", time):
        return track_set.as_cluster_time_stack(time)

    def __len__(self):
        return len(self.time)

class Track(_Geometry):
    """
    A track is  the drift track of a single buoy in time
    """

    def __init__(self, points: List[SpaceTimePoint], id):
        points = list(filter( lambda x:x.is_valid, points))
        self.points = points
        self.id = id

    @property
    def time(self) -> numpy.ndarray:
        return to_datetime64([x.time for x in self.points])

    @property
    def latitude(self) -> numpy.ndarray:
        return numpy.array([x.latitude for x in self.points])

    @property
    def longitude(self) -> numpy.ndarray:
        return numpy.array([x.longitude for x in self.points])

    def interpolate(self, target_time) -> "Track":
        target_time = to_datetime64(target_time)
        time = self.time
        latitude  = interpolate_periodic(
            time.astype('float64'),
            self.latitude,
            target_time.astype('float64'))
        longitude = interpolate_periodic(
            time.astype('float64'),
            self.longitude,
            target_time.astype('float64'),
            fp_period=360)
        return self.from_arrays(latitude,longitude,target_time,self.id)

    @classmethod
    def from_arrays(cls, latitude,longitude,time,id) -> "Track":
        points = []
        for lat,lon,t in zip(latitude,longitude,time):
            points.append(SpaceTimePoint(lat,lon,id,to_datetime_utc(t)) )
        return cls(points,id)

    @classmethod
    def from_spotter(cls,spotter_id,spotter):
        points = []
        if isinstance(spotter, DataFrame):
            for latitude,longitude,time in zip( spotter['latitude'].values,
                                                spotter['longitude'].values,
                                                spotter.index.values):
                points.append(SpaceTimePoint(
                    time=time,
                    latitude=latitude,
                    longitude=longitude,
                    id=spotter_id
                    )
                )
        elif hasattr(spotter, 'dataset'):
            for latitude,longitude,time in zip( spotter.dataset['latitude'].values,
                                                spotter.dataset['longitude'].values,
                                                spotter.dataset['time'].values):
                points.append(SpaceTimePoint(
                    time=time,
                    latitude=latitude,
                    longitude=longitude,
                    id=spotter_id
                    )
                )
        else:
            for x in spotter:
                points.append(SpaceTimePoint(
                    time=x.time,
                    latitude=x.latitude,
                    longitude=x.longitude,
                    id = spotter_id
                    )
                )
        return cls( points, spotter_id )

    def __len__(self):
        return len(self.points)

@dataclass()
class TrackSet(_Geometry):
    """
    A collection of tracks is a set of tracks for multiple buoys.
    """
    tracks: Mapping[str,Track]

    @classmethod
    def from_spotters(cls,spotters:Mapping):
        tracks = {}
        for spotter_id,spotter in spotters.items():
            tracks[spotter_id] = Track.from_spotter(spotter_id,spotter)
        return cls(tracks)

    @classmethod
    def from_clusters(cls,cluster_time_stack:"ClusterStack"):
        tracks = {}
        for cluster,time in zip(cluster_time_stack.clusters,
                                cluster_time_stack.time):

            for _id,point in cluster.points:
                space_time_point = SpaceTimePoint(
                    longitude=point.latitude,
                    latitude=point.longitude,
                    time=time,
                    id=_id
                )
                if _id not in tracks:
                    tracks[_id] = Track([space_time_point],_id)
                else:
                    tracks[_id].points.append(space_time_point)
        return cls(tracks)

    def as_cluster_time_stack(self, time):
        tracks = self.interpolate(time)

        clusters = [ {} for x in range(len(time)) ]
        for _id,track in tracks.tracks.items():
            for ii, point in enumerate(track.points):
                if point.is_valid:
                    index = numpy.searchsorted(
                        time, [to_datetime64(point.time)],side='left')[0]

                    clusters[index][_id] = SpatialPoint(
                        point.latitude,
                        point.longitude,
                        _id
                    )

        clusters = [ Cluster(x) for x in clusters ]
        return ClusterStack(time,clusters)

    def interpolate(self, time) -> "TrackSet":
        time = to_datetime64(time)
        tracks = {}
        for id,track in self.tracks.items():
            tracks[id] = track.interpolate(time)
        return TrackSet(tracks)

    @classmethod
    def from_cluster(cls,cluster:Cluster,
                         time: numpy.ndarray) -> "TrackSet":
        tracks = {}
        for _id, point in cluster.points.items():
            track_points = [
                SpaceTimePoint.from_spatial_point(point,t) for t in time ]
            tracks[point.id] = Track( points=track_points, id=point.id)
        return TrackSet(tracks)

Geometry = Union[_Geometry, Sequence[numpy.ndarray],Sequence[Sequence],
                            Sequence[Number],Mapping]

def convert_to_cluster_stack(geometry:Geometry, time:numpy.ndarray
                             ) -> ClusterStack:

    if isinstance(geometry, TrackSet):
        return geometry.as_cluster_time_stack(time)

    elif isinstance(geometry, Cluster):
        return ClusterStack(time, [ geometry for _ in time ])

    elif isinstance(geometry, Track):
        tracks = TrackSet({'track':geometry})
        return ClusterStack.from_track_set( tracks,time)

    elif isinstance(geometry, ClusterStack):
        return geometry

    elif isinstance(geometry, Sequence):
        tracks = convert_to_track_set(geometry,time)
        return ClusterStack.from_track_set( tracks,time)

    elif isinstance(geometry, Mapping):
        tracks = convert_to_track_set(geometry,time)
        return ClusterStack.from_track_set(tracks, time)


def convert_to_track_set(geometry:Geometry, time:numpy.ndarray
                         ) -> TrackSet:

    if isinstance(geometry, TrackSet):
        return geometry.interpolate(time)

    elif isinstance(geometry, Cluster):
        return TrackSet.from_clusters(
            ClusterStack(time, [ geometry for _ in time ]))

    elif isinstance(geometry, Track):
        return TrackSet({'track':geometry.interpolate(time)})

    elif isinstance(geometry, ClusterStack):
        return geometry.as_track_set()

    elif isinstance(geometry, Sequence):
        if not (isinstance(geometry[0],Sequence) or
                            isinstance(geometry[0],Sequence)):
            _geometry = [ [geometry[0]] , [geometry[1]]  ]
        else:
            _geometry = geometry

        cluster = Cluster.from_lat_lon_arrays(
                _geometry[0],_geometry[1])
        return TrackSet.from_cluster(cluster, time)

    elif isinstance(geometry,Mapping):
        return TrackSet.from_spotters(geometry).interpolate(time)





from pandas import DataFrame
from roguewave.interpolate.general import interpolate_periodic
from roguewave.tools.time import to_datetime64, to_datetime_utc
import numpy

def interpolate_dataframe_time( dataframe:DataFrame, new_time:numpy.ndarray):

    output = DataFrame(index=to_datetime_utc(new_time))
    columns = list(dataframe.columns)
    old_time = to_datetime64(dataframe.index.values)
    new_time = to_datetime64(new_time)


    # Todo more general direction detection.

    for name in columns:
        name: str
        period = None
        if 'direction' in name.lower():
            fp_discont = 360
            fp_period  = 360
        else:
            fp_discont = None
            fp_period  = None

        output[name] = interpolate_periodic(
            old_time.astype('float64'),
            dataframe[name].values,
            new_time.astype('float64'),
            x_period=period,
            fp_period=fp_period,
            fp_discont=fp_discont
        )
    return output


"""
Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""

from datetime import datetime, timezone
from xarray import DataArray
import numpy
from typing import Union, List, Tuple
from numpy import datetime64

scalar_input_types = Union[float, int, datetime, str, datetime64]
input_types = Union[scalar_input_types,List[scalar_input_types],numpy.ndarray]

def to_datetime_utc(time: input_types,to_scalar=False
                    ) -> Union[datetime, List[datetime]]:
    datetimes = _to_datetime_utc(time)
    if to_scalar and not isinstance(datetimes,datetime):
        return datetimes[0]
    else:
        return datetimes

def _to_datetime_utc(time: input_types) -> Union[datetime, List[datetime]]:
    if time is None:
        return None

    if isinstance(time, datetime):
        return time.astimezone(timezone.utc)

    elif isinstance(time, str):
        if time[-1] == "Z":
            # From isoformat does not parse "Z" as a valid timezone designator
            time = time[:-1] + "+00:00"
        try:
            return datetime.fromisoformat(time)
        except ValueError as e:
            return datetime.strptime(time, "%Y-%m-%dT%H:%M:%S.%f%z")

    elif isinstance(time,datetime64):
        time_stamp_seconds = numpy.datetime64(time,'s').astype('float64')
        time =  datetime.fromtimestamp( time_stamp_seconds ,tz=timezone.utc )
        return time

    elif isinstance(time,List) or isinstance(time,numpy.ndarray):
        return [ to_datetime_utc(x) for x in time ]

    elif isinstance(time,DataArray):
        time = time['time'].values
        return [ to_datetime_utc(x) for x in time ]

    else:
        return datetime.fromtimestamp(time, tz=timezone.utc)

def datetime_to_iso_time_string(time: input_types):
    if time is None:
        return None

    time = to_datetime_utc(time)
    return time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

def to_datetime64( time:Union[input_types,numpy.array,List,Tuple]
                   ) -> Union[numpy.ndarray,None]:
    """
    Convert time input to numpy ndarrays.
    :param time:
    :return:
    """
    if time is None:
        return None

    if isinstance(time, numpy.ndarray):
        return numpy.array([datetime64(to_datetime_utc(x),'s') for x in time])

    elif isinstance( time, List ) or isinstance( time, Tuple ):
        return numpy.array([datetime64(to_datetime_utc(x),'s') for x in time])

    elif isinstance( time, datetime64 ):
        return datetime64(time,'s')
    elif isinstance(time, datetime):
        return datetime64( int(time.timestamp()),'s')
    else:
        raise ValueError('unknown time type')
